from typing import Union

from CosmoTech_Acceleration_Library.Accelerators.adx_wrapper import ADXQueriesWrapper
from Supplychain.Generic.adx_wrapper import ADXWrapper
from Supplychain.Schema.adt_column_description import ADTColumnDescription
from Supplychain.Generic.timer import Timer
from Supplychain.Generic.folder_io import FolderReader, FolderWriter

from Supplychain.Generic.adx_and_file_writer import ADXAndFileWriter

dataset_description = ADTColumnDescription.format


class FromDictToTableConverter(Timer, ADXAndFileWriter):

    def max_time_step(self,
                      file_name: str) -> int:
        file_to_convert = self.reader.files.get(file_name, [])
        file_description = dataset_description[file_name]

        time_elements = list(file_description['change'])

        max_time_step = 0

        for events in file_description['event'].values():
            time_elements.extend(events)

        for element in file_to_convert:
            _local_max = 0
            for time_element in time_elements:
                time_dict = element.get(time_element, {})
                if time_dict:
                    _end_time_step = max(map(int, time_dict.keys()))
                    _local_max = max(_local_max, _end_time_step)
            max_time_step = max(max_time_step, _local_max)
        return max_time_step

    def convert_file(self,
                     file_name: str,
                     base_row_keys: tuple):
        """
        Will open a given CSV file and convert it to a better ADX format
        The file will be separated into 2 files :
        - a "fixed" file containing non changing data
        - a "temporal" file containing all the temporal data associated to every time step by extending the data
        :param file_name: the file to convert
        :param base_row_keys: a list of rows from the fixed data that should be used as index for the temporal
        :return: None
        """

        file_to_convert = self.reader.files.get(file_name, [])

        fixed_data_filename = file_name
        temporal_data_filename = file_name + "Schedules"

        file_description = dataset_description[file_name]

        fixed_data = list()
        temporal_data = list()

        events_filenames = list(file_description['event'].keys())
        event_data = dict()
        for _name in events_filenames:
            event_data[_name] = list()

        for element in file_to_convert:
            # Each element will be converted separatly
            current_fixed_data = dict()
            if self.simulation_id is not None:
                current_fixed_data['SimulationRun'] = self.simulation_id
            current_events = dict()

            # First fixed data are read
            for _key in file_description['fixed']:
                if _key in element:
                    current_fixed_data[_key] = element[_key]
                else:
                    current_fixed_data[_key] = None

            def read_data(column_list: list) -> dict:
                schedule = dict()
                for change in column_list:
                    value = element.get(change, None)
                    if value:
                        _timesteps = sorted(value.keys(), key=int)
                        # We create our list of time steps
                        for _timestep in _timesteps:
                            if _timestep not in schedule:
                                schedule[_timestep] = dict()
                            if _timestep not in value:
                                continue
                            measure = value[_timestep]
                            # This time we have a single key so we directly add it's value
                            schedule[_timestep][change] = measure
                return schedule

            # We read the temporal data for "changes" and "events"
            current_changes = read_data(file_description['change'])
            for event_name, event_columns in file_description['event'].items():
                current_events[event_name] = read_data(event_columns)

            fixed_data.append(current_fixed_data)

            # We prepare a data structure to keep the latest change value for our data
            current_change_value = dict()
            for k in file_description['change']:
                # Needs a default value depending on the key
                current_change_value[k] = 0

            # We create a base row which will be duplicated and filed for each timestep
            present_base = [_k for _k in base_row_keys if _k in current_fixed_data]
            base_row = {_k: current_fixed_data[_k] for _k in present_base}
            previous_row = None
            for timestep in range(self.total_time_step):
                if file_description['change']:
                    current_row = base_row.copy()
                    current_row['Timestep'] = timestep

                    def change_from_key(local_key):
                        # In case of a change if the value exists in the data at a given time step
                        # we update our current value then set it in the row
                        if timestep in current_changes and local_key in current_changes[timestep]:
                            current_change_value[local_key] = current_changes[timestep][local_key]
                        current_row[local_key] = current_change_value[local_key]

                    # We loop over all our keys in "changes" and in "events"
                    for change_key in file_description['change']:
                        change_from_key(change_key)

                    # If we have any data in "changes" or "events" we keep the created row
                    to_be_kept = True
                    if previous_row is not None and not self.keep_duplicate:
                        to_be_kept = False
                        for k in previous_row.keys():
                            if k != "Timestep":
                                if current_row[k] != previous_row[k]:
                                    to_be_kept = True
                                    break
                    if to_be_kept:
                        previous_row = current_row
                        temporal_data.append(current_row)

                for event_name, event_columns in file_description['event'].items():
                    event_row = base_row.copy()
                    event_row['Timestep'] = timestep
                    event_exists = False
                    current_event = current_events[event_name]
                    for event_key in event_columns:
                        if timestep in current_event and event_key in current_event[timestep]:
                            event_row[event_key] = current_event[timestep][event_key]
                            event_exists = True
                        else:
                            # Needs a default value depending on the key
                            event_row[event_key] = 0.0
                    if event_exists or self.keep_duplicate:
                        event_data[event_name].append(event_row)

        if fixed_data:
            self.write_target_file(fixed_data, fixed_data_filename, self.simulation_id)
        else:
            fake_fixed_element = dict()
            for key in file_description['fixed']:
                fake_fixed_element[key] = None
            if self.simulation_id:
                fake_fixed_element['SimulationRun'] = self.simulation_id
            self.write_target_file([fake_fixed_element, ], fixed_data_filename, self.simulation_id)

        fake_row_keys = base_row_keys
        if self.simulation_id is None and 'SimulationRun' in base_row_keys:
            fake_row_keys = [k for k in base_row_keys if k != 'SimulationRun']

        if temporal_data:
            self.write_target_file(temporal_data, temporal_data_filename, self.simulation_id)
        elif file_description['change']:
            fake_temporal_element = dict()
            for column_list in fake_row_keys, ['Timestep', ], file_description['change']:
                for key in column_list:
                    fake_temporal_element[key] = None
            if self.simulation_id:
                fake_temporal_element['SimulationRun'] = self.simulation_id
            self.write_target_file([fake_temporal_element, ], temporal_data_filename, self.simulation_id)
        for event_name, event_file_content in event_data.items():
            if event_file_content:
                self.write_target_file(event_file_content, event_name, self.simulation_id)
            elif file_description['event'][event_name]:
                fake_event_element = dict()
                for column_list in fake_row_keys, ['Timestep', ], file_description['event'][event_name]:
                    for key in column_list:
                        fake_event_element[key] = None
                if self.simulation_id:
                    fake_event_element['SimulationRun'] = self.simulation_id
                self.write_target_file([fake_event_element, ], event_name, self.simulation_id)
        self.split("{time_since_last_split:6.4f}({time_since_start:6.4f}) - " + file_name)

    def __init__(self,
                 simulation_id: Union[str, None],
                 reader: FolderReader,
                 writer: Union[FolderWriter, None] = None,
                 adx_connector: Union[ADXQueriesWrapper, ADXWrapper, None] = None,
                 keep_duplicate: bool = True):
        """
        Init and execute transformation of the data
        :param reader: Folder reader serving files
        :param writer: Potential folder writer
        :param simulation_id: ID of the simulation to be added to every row
        :param adx_connector: Potential connecto to ADX
        :param keep_duplicate: should duplicate rows be kept ?
        """

        # Initialize the ADX wrapping
        Timer.__init__(self, '[Save to ADX]')
        ADXAndFileWriter.__init__(self,
                                  writer=writer,
                                  adx_connector=adx_connector)

        self.simulation_id = simulation_id

        self.keep_duplicate = keep_duplicate

        self.reader = reader

        # Get the total of time_steps from the configuration file
        self.total_time_step = 0

    def convert(self):
        # lookup for max time step
        if 'Configuration' in self.reader.files and len(self.reader.files['Configuration']) > 0:
            self.total_time_step = (self.reader.files['Configuration'][0]['SimulatedCycles']
                                    * self.reader.files['Configuration'][0]['StepsPerCycle'])
            self.convert_file("Configuration", ())
        else:
            self.total_time_step = max(map(self.max_time_step, ["ProductionResource",
                                                                "ProductionOperation",
                                                                "Stock",
                                                                "Transport", ])) + 1

        self.convert_file("ProductionResource", ("SimulationRun", "id",))
        self.convert_file("ProductionOperation", ("SimulationRun", "id",))
        self.convert_file("Stock", ("SimulationRun", "id",))

        self.convert_file("contains", ())
        self.convert_file("input", ())
        self.convert_file("output", ())
        self.convert_file("Transport", ("SimulationRun",
                                        "Label",))  # TODO to be replaced by RelationshipId when available from azure-digital-twins-simulator-connector



from typing import Union
from Supplychain.Generic.folder_io import FolderWriter, FolderReader


class MemoryFolderIO(FolderWriter, FolderReader):

    def reset(self):
        self.files = dict()

    def write_from_list(self, dict_list: list, file_name: str, ordering_key: Union[str, None] = None):
        if dict_list:
            to_be_writen = dict_list
            if ordering_key is not None and ordering_key in dict_list[0]:
                to_be_writen = sorted(dict_list, key=lambda e: e[ordering_key])
            self.files[file_name] = to_be_writen

    def __init__(self):

        FolderReader.__init__(self)
        FolderWriter.__init__(self)

        self.reset()



from Supplychain.Generic.folder_io import FolderReader
from Supplychain.Generic.timer import Timer
from Supplychain.Schema.validation_schemas import ValidationSchemas
from Supplychain.Schema.default_values import parameters_default_values

from collections import Counter

import itertools
import jsonschema


class DictValidator(Timer):

    def __init__(self,
                 reader: FolderReader,
                 run_type: str = "None"):
        Timer.__init__(self, prefix="[Validation]")

        self.reader = reader

        self.schema = ValidationSchemas()

        self.run_type = run_type

        self.errors = []

        self.lookup_memory = dict()

    def validate(self) -> bool:
        checks = [self.validate_files(), self.validate_graph(), self.specific_validations()]
        if all(checks):
            self.display_message("Dataset is valid")
            return True
        else:
            self.display_message("Dataset is invalid")
            return False

    def validate_graph(self) -> bool:
        self.errors = []
        self.display_message("Validate graph")
        for file_to_validate in sorted(self.schema.graph.keys()):
            self.__validate_graph(file_to_validate)
        if self.errors:
            self.display_message(f"{len(self.errors)} error{'s' if len(self.errors) > 1 else ''} found in the files")
            for filename, item_id, err in self.errors:
                self.display_message(f" - {filename}: {item_id}: {err}")
            return False
        self.__validate_acyclicity()
        if self.errors:
            self.display_message("1 error found in the files")
            self.display_message(self.errors[0])
            return False
        self.display_message("Graph is valid")
        return True

    def validate_files(self) -> bool:
        self.errors = []
        self.display_message("Validate file content")
        for file_to_validate in sorted(self.schema.schemas.keys()):
            self.__validate_file(file_to_validate)

        if self.errors:
            self.display_message(f"{len(self.errors)} error{'s' if len(self.errors) > 1 else ''} found in the files")
            for filename, item_id, err in self.errors:
                message = err.message.replace("{", "{{").replace("}", "}}")
                self.display_message(f" - {filename}: {item_id}: {'/'.join(err.path)}: {message}")
        else:
            self.display_message("Individual files are valid")
            return True
        return False

    def __validate_file(self, file_name: str):
        for item in self.reader.files[file_name]:
            validator = jsonschema.Draft7Validator(schema=self.schema.schemas[file_name])
            errors = validator.iter_errors(item)
            item_id = item['id'] if 'id' in item else item.get("Label",
                                                               None)  # TODO specific for relationship here transport (only Transport relationship is validated) to be replaced by RelationshipId when available from azure-digital-twins-simulator-connector
            if item_id is None:
                item_id = item.get('source', "") + " -> " + item.get('target', "")
            for error in errors:
                self.errors.append((file_name, item_id, error))
        self.split(f"\t{file_name}" + ": {time_since_last_split:6.4f}s")

    def __validate_graph(self, file_name: str):
        config = self.schema.graph[file_name]
        source = config['source']
        target = config['target']
        source_file, source_id = config['links']['source']
        target_file, target_id = config['links']['target']
        cardinality_target, cardinality_source = config['cardinalities'].split(':')
        all_target = config['all_target_present']
        all_source = config['all_source_present']

        source_file_ids = [i[source_id] for i in self.reader.files[source_file]]
        target_file_ids = [i[target_id] for i in self.reader.files[target_file]]

        arcs = [(i[source], i[target]) for i in self.reader.files[file_name]]
        if arcs:
            sources, targets = zip(*arcs)

            # Check required existences
            for check, file_ids, ids in [(all_source, source_file_ids, sources),
                                         (all_target, target_file_ids, targets)]:
                if check:
                    for item_id in file_ids:
                        if item_id not in ids:
                            self.errors.append((file_name, item_id, "has no relations"))

            for file_ids, ids, current_file in [(source_file_ids, sources, source_file),
                                                (target_file_ids, targets, target_file)]:
                for item_id in ids:
                    if item_id not in file_ids:
                        self.errors.append((file_name, item_id, f"does not exist in {current_file}"))

            # Check cardinalities
            for entities, cardinality in [(sources, cardinality_source), (targets, cardinality_target)]:
                for e in set(entities):
                    if cardinality == "1" and entities.count(e) > 1:
                        self.errors.append((file_name, e, "has more than one relation"))

        self.split(f"\t{file_name}" + ": {time_since_last_split:6.4f}s")

    def __validate_acyclicity(self):
        arcs_by_file_name = {
            file_name: [
                (i[config['source']], i[config['target']])
                for i in self.reader.files[file_name]
            ]
            for file_name, config in self.schema.graph.items()
        }
        arcs = [
            arc
            for arcs_of_file in arcs_by_file_name.values()
            for arc in arcs_of_file
        ]
        if arcs:
            vertices = set(itertools.chain(*arcs))
            targets_by_source = {
                vertex: set()
                for vertex in vertices
            }
            for source, target in arcs:
                targets_by_source[source].add(target)
            visited = set()
            loop = []

            def visit(vertex):
                if vertex in visited:
                    return False
                if vertex in loop:
                    loop.append(vertex)
                    return True
                vertices.discard(vertex)
                loop.append(vertex)
                for next_vertex in targets_by_source[vertex]:
                    if visit(next_vertex):
                        return True
                visited.add(loop.pop())

            while vertices:
                if visit(vertices.pop()):
                    break
            if loop:
                loop = loop[loop.index(loop[-1]):]
                vertex_types = []
                file_names = []
                for arc in zip(loop[:-1], loop[1:]):
                    for file_name, arcs_of_file in arcs_by_file_name.items():
                        if arc in arcs_of_file:
                            break
                    vertex_types.append(self.schema.graph[file_name]['links']['source'][0])
                    file_names.append(file_name)
                spacing = max(
                    max(map(len, vertex_types)),
                    max(map(len, file_names)) - 2,
                )
                vertex_types.append(vertex_types[0])
                sep = '\n\t'
                relations = [sep] + [
                    f'{file_name:>{spacing + 2}} ↓{sep}'
                    for file_name in file_names
                ]
                loop_links = [
                    f"{relation}[{vertex_type:^{spacing}}] {vertex}"
                    for relation, vertex_type, vertex in zip(relations, vertex_types, loop)
                ]
                self.errors.append(f"The graph contains at least one loop:{sep.join(loop_links)} (same as first)")

    def specific_validations(self) -> bool:
        # If specific validations are required add them here
        self.errors = []
        self.display_message("Specific validations")

        self.__transports_specific_checks()
        self.split("\ttransports_specific_checks: {time_since_last_split:6.4f}s")
        for filename, id_column in [('Transport', 'Label'),  # TODO replace by
                                    ('ProductionOperation', 'id'),
                                    ('ProductionResource', 'id'),
                                    ('Stock', 'id')]:
            self.__unique_id_validation(filename, id_column)
        self.split("\tunique_id_validation: {time_since_last_split:6.4f}s")

        if self.run_type != "Simulation":
            self.__part_retention_validation()
            self.split("\tpart_retention_validation: {time_since_last_split:6.4f}s")

        self.__no_infinite_stocks_to_transport()
        self.split("\tno_infinite_stocks_to_transport: {time_since_last_split:6.4f}s")

        self.__obsolescence_check()
        self.split("\tobsolescence_check: {time_since_last_split:6.4f}s")

        self.__check_resource_dependencies()
        self.split("\tcheck_resource_dependencies: {time_since_last_split:6.4f}s")

        self.__mandatory_attributes_check()
        self.split("\tmandatory_attributes_check: {time_since_last_split:6.4f}s")

        if self.errors:
            self.display_message(f"{len(self.errors)} error{'s' if len(self.errors) > 1 else ''} found in the files")
            for filename, item_id, err in self.errors:
                self.display_message(f" - {filename}: {item_id}: {err}")
        else:
            self.display_message("Specific checks are valid")
            return True
        return False

    def __part_retention_validation(self):
        stocks = self.reader.files['Stock']
        transports = self.reader.files['Transport']
        outputs = self.reader.files['input']
        isdp = None
        if self.reader.files['Configuration']:
            isdp = self.reader.files['Configuration'][0].get('IntermediaryStockDispatchPolicy')
        if isdp is None:
            isdp = parameters_default_values['Configuration']['IntermediaryStockDispatchPolicy']
        if isdp == 'DispatchAll':
            non_final_stocks = set()
            for t in transports:
                non_final_stocks.add(t['source'])
            for o in outputs:
                non_final_stocks.add(o['source'])
            for stock in stocks:
                if stock['id'] not in non_final_stocks:
                    continue
                stock_demands = stock.get('Demands')
                if stock_demands is None:
                    continue
                if any(stock_demands.values()):
                    self.errors.append(("Stock",
                                        stock['id'],
                                        "has demands and is not a final stock,"
                                        " set IntermediaryStockDispatchPolicy to AllowRetention"))

    def __unique_id_validation(self, filename: str, id_column: str):
        _file = self.reader.files[filename]

        _ids = Counter(item[id_column] for item in _file)

        for _id, count in _ids.items():
            if count > 1:
                self.errors.append((filename,
                                    _id,
                                    f"{id_column} is not unique, found {count} times"))

    def __transports_specific_checks(self):
        transports = self.reader.files['Transport']
        for transport in transports:
            duration = transport.get('Duration')
            if 'InitialTransportQuantities' in transport:
                if duration is not None:
                    key_errors = [k for k in transport['InitialTransportedQuantities'] if int(k) >= duration]
                    for k in key_errors:
                        self.errors.append(("Transports",
                                            transport.get('Label'),
                                            f"InitialTransportedQuantities: key {k} >= {duration}"))
            if 'InitialTransportValues' in transport:
                if duration is not None:
                    key_errors = [k for k in transport['InitialTransportedValues'] if int(k) >= duration]
                    for k in key_errors:
                        self.errors.append(("Transports",
                                            transport.get('Label'),
                                            f"InitialTransportedValues: key {k} >= {duration}"))
                key_errors = [k
                              for k in transport['InitialTransportedValues']
                              if k not in transport['InitialTransportedQuantities']]
                for k in key_errors:
                    self.errors.append(("Transports",
                                        transport.get('Label'),
                                        f"InitialTransportedValues: key {k} has no quantities associated"))

    def __no_infinite_stocks_to_transport(self):
        transports = self.reader.files['Transport']
        stocks = self.reader.files['Stock']

        infinite_stocks = [stock['id'] for stock in stocks if stock.get('IsInfinite')]

        transports_from_infinite_stocks = [transport
                                           for transport in transports
                                           if transport['source'] in infinite_stocks]

        for transport in transports_from_infinite_stocks:
            self.errors.append(("Stock",
                                transport.get('source'),
                                f"Is infinite and has outgoing transport: {transport.get('Label')}"))

    def __obsolescence_check(self):
        empty_obsolete_stocks = None
        if self.reader.files['Configuration']:
            empty_obsolete_stocks = self.reader.files['Configuration'][0].get('EmptyObsoleteStocks')
        if empty_obsolete_stocks is None:
            empty_obsolete_stocks = parameters_default_values['Configuration']['EmptyObsoleteStocks']
        if empty_obsolete_stocks:
            stocks = self.reader.files['Stock']
            for stock in stocks:
                stock_demands = stock.get('Demands')
                if stock_demands is None:
                    continue
                has_demands = any(stock_demands.values())
                stock_policy = stock.get('StockPolicy')
                if stock_policy is None:
                    stock_policy = parameters_default_values['Configuration']['StockPolicy']
                has_stock_policy = stock_policy != 'None'
                if has_demands and has_stock_policy:
                    self.errors.append(('Configuration',
                                        'EmptyObsoleteStocks',
                                        'The stock obsolescence option is not compatible with stock policies.'))
                    break


    def find_relations_by_id(self, relation_name: str, looked_id: str, relation_column: str):
        if (relation_name, looked_id, relation_column) not in self.lookup_memory:
            self.lookup_memory[(relation_name, looked_id, relation_column)] = [
                row
                for row in self.reader.files[relation_name]
                if row.get(relation_column) == looked_id
            ]
        return self.lookup_memory[(relation_name, looked_id, relation_column)]

    parent_operations_mem = dict()

    def __find_parent_operations(self, stock_id: str):
        if stock_id in self.parent_operations_mem:
            return self.parent_operations_mem[stock_id]

        _ret = []
        self.parent_operations_mem[stock_id] = _ret
        # transports won't increment the current level
        for _transport in self.find_relations_by_id('Transport', stock_id, 'target'):
            for operation in self.__find_parent_operations(_transport.get('source')):
                _ret.append(operation)

        for _output in self.find_relations_by_id('output', stock_id, 'target'):
            _operation_id = _output.get('source')
            _ret.append(_operation_id)
        return _ret


    def __check_resource_dependencies(self):
        # A resource should not have a "loop" dependency
        # It means if any operation in the resource is dependant on another operation in the same resource
        # the dataset validation should fail
        # A dependency is either a direct parent operation, or an operation in the same resource as a parent

        contains = self.reader.files['contains']

        direct_dependencies = dict()
        operations_per_resource = dict()
        resource_per_operation = dict()

        # fill dependencies
        for contain in contains:
            operation_id = contain.get('target')
            resource_id = contain.get('source')
            operations_per_resource.setdefault(resource_id, [])
            operations_per_resource[resource_id].append(operation_id)
            resource_per_operation[operation_id] = resource_id
            direct_dependencies.setdefault(operation_id, set())
            for _input in self.find_relations_by_id('input', operation_id, 'target'):
                direct_dependencies[operation_id].update(self.__find_parent_operations(_input.get('source')))

        for operations in operations_per_resource.values():
            dependencies = set()
            for operation_id in operations:
                dependencies.update(direct_dependencies[operation_id])
                direct_dependencies[operation_id] = dependencies

        full_dependencies = dict()

        def find_dependencies(operation_id: str):
            if operation_id in full_dependencies:
                return full_dependencies[operation_id]
            _ret = set()
            full_dependencies[operation_id] = _ret
            _dependencies = direct_dependencies.get(operation_id, set())
            for dependency in _dependencies:
                _ret.add(dependency)
                _ret.update(find_dependencies(dependency))
            return _ret

        for operation_id, resource_id in resource_per_operation.items():
            if operation_id in find_dependencies(operation_id):
                self.errors.append(('Graph',
                                    'Resource dependency',
                                    f'In resource {resource_id}, the operation {operation_id} requires itself'))
                return

    def __mandatory_attributes_check(self):
        demand = any(
            len(stock_demands.values())
            for stock in self.reader.files['Stock']
            for stock_demands in (stock.get('Demands'),)
            if stock_demands
        )
        if not demand:
            self.errors.append(('Stock',
                                'Demands',
                                'There is no demand.'))
        # for future warnings:
        """
        for resource in self.reader.files['ProductionResource']:
            if 'OpeningTimes' not in resource:
                self.display_message(f"Production resource {resource['id']} has no opening time.")
        for operation in self.reader.files['ProductionOperation']:
            if 'CycleTimes' not in operation:
                self.display_message(f"Production operation {operation['id']} has no opening time.")
        for transport in self.reader.files['Transport']:
            if 'Duration' not in transport:
                self.display_message(f"Transport {transport['Label']} has no duration.")
        """



from Supplychain.Generic.cosmo_api_parameters import CosmoAPIParameters
from Supplychain.Generic.folder_io import FolderWriter, FolderReader
from Supplychain.Wrappers.environment_variables import SECONDS_IN_MINUTE
from Supplychain.Generic.excel_folder_reader import ExcelReader
from Supplychain.Generic.csv_folder_reader import CSVReader
from Supplychain.Generic.memory_folder_io import MemoryFolderIO
from Supplychain.Transform.from_table_to_dict import write_transformed_data
from Supplychain.Generic.timer import Timer
from dateutil import parser
import math
from typing import Union

GRANULARITIES = {
    'minute':
        1,
    'hour':
        60,
    'day':
        60 * 24,
    'week':
        60 * 24 * 7,
    'month':
        60 * 24 * 30,
    'quarter':
        60 * 24 * 90,
    'year':
        60 * 24 * 365
}


def object_to_bool(item: Union[object, str]) -> bool:
    if type(item) is str:
        return item.lower() == 'true'
    return bool(item)


class DictPatcher(Timer):

    def __write_updated_files(self):
        to_use_reader = self.memory

        # ADT column names to be replaced for post ADT simulation
        name_replacements = {"$sourceId": "source",
                             "$targetId": "target",
                             "$id": "id"}

        for file_name, file_content in to_use_reader.files.items():
            self.writer.write_from_list(dict_list=[
                {
                    name_replacements.get(item_k, item_k): item_v
                    for item_k, item_v in item.items()
                }
                for item in file_content], file_name=file_name)

    def __init__(self,
                 reader: FolderReader,
                 writer: FolderWriter,
                 parameters: CosmoAPIParameters):
        Timer.__init__(self, "[ParameterHandler]")
        self.parameters = parameters
        self.writer = writer
        self.reader = reader

        self.memory = MemoryFolderIO()
        for file_name, file_content in self.reader.files.items():
            self.memory.write_from_list(dict_list=file_content, file_name=file_name)
        if 'Configuration' not in self.memory.files:
            self.memory.files['Configuration'] = [dict(),]

    def handle_mass_action_lever(self):
        try:
            lever_folder = self.parameters.get_dataset_path("mass_lever_excel_file")
        except ValueError:
            self.display_message("No mass action lever found - skipping")
            return False, 0
        lever_reader = ExcelReader(lever_folder, keep_nones=False)
        self.memory.reset()

        reading_errors = write_transformed_data(
            reader=lever_reader,
            writer=self.memory
        )
        if reading_errors:
            return True, reading_errors
        if not self.memory.files.setdefault('Configuration', []):
            self.memory.files['Configuration'].append({})

        self.__write_updated_files()
        return True, 0

    def handle_simple_simulation(self):
        configuration = self.memory.files['Configuration'][0]

        start_date = parser.isoparse(self.parameters.get_named_parameter('start_date').value)
        end_date = parser.isoparse(self.parameters.get_named_parameter('end_date').value)
        simulation_granularity = GRANULARITIES.get(self.parameters.get_named_parameter('simulation_granularity').value,
                                                   GRANULARITIES['day'])

        duration = (end_date - start_date).total_seconds() // SECONDS_IN_MINUTE

        steps = int(math.ceil(duration / simulation_granularity))

        self.display_message(f"Starting Date: {start_date.isoformat()}")
        self.display_message(f"Simulated Cycles: {steps}")
        self.display_message(f"Steps per Cycle: {1}")
        self.display_message(f"TimeStep Duration: {simulation_granularity} minutes")

        configuration['StartingDate'] = start_date.isoformat()
        configuration['SimulatedCycles'] = steps
        configuration['StepsPerCycle'] = 1
        configuration['TimeStepDuration'] = simulation_granularity

        self.__handle_demand_plan()
        self.__handle_transport_duration()
        self.__handle_production_resource_opening_time()

        self.__write_updated_files()

    def handle_optimization_parameter(self):
        parameters = (('optimization_objective', 'OptimizationObjective', 'Optimization Objective', None),)
        for name, config_name, display_name, type_f in parameters:
            self.__handle_configuration_parameter(name, config_name, display_name, type_f)
        self.__write_updated_files()

    def handle_flow_management_policies(self):
        parameters = (('stock_policy', 'Stock', 'StockPolicy'),
                      ('sourcing_policy', 'Stock', 'SourcingPolicy'),
                      ('stock_dispatch_policy', 'Stock', 'DispatchPolicy'),
                      ('production_policy', 'ProductionResource', 'ProductionPolicy'))
        for parameter_name, entity_type, entity_parameter_name in parameters:
            try:
                parameter_value = self.parameters.get_named_parameter(parameter_name).value
            except ValueError:
                self.display_message(f"{parameter_name} is not defined - skipping")
                continue
            if parameter_value == 'FromDataset':
                continue
            if entity_type not in self.memory.files:
                continue
            entities = self.memory.files[entity_type]
            for entity in entities:
                entity[entity_parameter_name] = parameter_value
            entities_count = len(entities)
            self.display_message(f'Set {entity_parameter_name} to {parameter_value} for all {entity_type} '
                                 f'({entities_count} entit{"ies" if entities_count > 1 else "y"})')
        self.__write_updated_files()

    def handle_model_behavior(self):
        parameters = (('manage_backlog_quantities', 'ManageBacklogQuantities', 'Manage Backlog Quantities',
                       object_to_bool),
                      ('empty_obsolete_stocks', 'EmptyObsoleteStocks', 'Empty Obsolete Stocks',
                       object_to_bool),
                      ('batch_size', 'BatchSize', 'Batch Size',
                       int),
                      ('financial_cost_of_stocks', 'FinancialCostOfStock', 'Financial Cost Of Stock',
                       float),
                      ('intermediary_stock_dispatch', 'IntermediaryStockDispatchPolicy',
                       'Intermediary Stock Dispatch Policy', None))
        for name, config_name, display_name, type_f in parameters:
            self.__handle_configuration_parameter(name, config_name, display_name, type_f)
        self.__write_updated_files()

    def handle_uncertainties_settings(self):
        if self.__handle_configuration_parameter('uncertainties_probability_distribution',
                                                 'UncertaintiesProbabilityDistribution',
                                                 'Uncertainties Probability Distribution',
                                                 None):
            self.__write_updated_files()

    def __handle_configuration_parameter(self,
                                         parameter_name: str,
                                         configuration_parameter_name: str,
                                         display_name: str,
                                         type_function) -> bool:
        configuration = self.memory.files['Configuration'][0]
        try:
            parameter_value = self.parameters.get_named_parameter(parameter_name).value
        except ValueError:
            self.display_message(f"{parameter_name} is not defined - skipping")
            return False
        param_value = (parameter_value
                       if type_function is None
                       else type_function(parameter_value))
        self.display_message(f"{display_name}: {param_value}")
        configuration[configuration_parameter_name] = param_value
        return True


    def __handle_demand_plan(self):
        try:
            demand_plan_folder = self.parameters.get_dataset_path("demand_plan")
        except ValueError:
            self.display_message("No demand plan found - skipping")
            return
        reader = CSVReader(demand_plan_folder)
        stock_changes = dict()
        for k, demand_plan in reader.files.items():
            for _demand in demand_plan:
                stock_id = _demand.get('id')
                timestep = str(_demand.get('Timestep'))
                stock_changes.setdefault(stock_id, dict())
                for key in ['Demands', 'DemandUncertainties', 'DemandWeights']:
                    stock_changes[stock_id].setdefault(key, dict())
                    if key in _demand:
                        stock_changes[stock_id][key][timestep] = _demand.get(key)

        stocks = self.memory.files['Stock']
        for stock in stocks:
            _id = stock['id']
            for key in ['Demands', 'DemandUncertainties', 'DemandWeights']:
                if _id in stock_changes:
                    stock[key] = stock_changes[_id][key]
                else:
                    stock[key] = {}

    def __handle_transport_duration(self):
        try:
            transport_duration_folder = self.parameters.get_dataset_path("transport_duration")
        except ValueError:
            self.display_message("No lever for transport duration found - skipping")
            return
        reader = CSVReader(transport_duration_folder)

        duration_changes = dict()
        for k, transport_durations in reader.files.items():
            for _td in transport_durations:
                source = _td.get('source')
                target = _td.get('target')
                duration = int(_td.get('Duration'))
                duration_changes[(source, target)] = duration
        for e in self.memory.files['Transport']:
            duration = duration_changes.get((e.get('source'), e.get('target')))
            if duration:
                e['Duration'] = duration

    def __handle_production_resource_opening_time(self):
        try:
            pr_opening_time_folder = self.parameters.get_dataset_path("production_resource_opening_time")
        except ValueError:
            self.display_message("No lever for production resource opening time found - skipping")
            return
        reader = CSVReader(pr_opening_time_folder)

        opening_changes = dict()
        for k, changes in reader.files.items():
            for _td in changes:
                _id = _td.get('id')
                timestep = _td.get('Timestep')
                opening_time = float(_td.get('OpeningTimes'))
                opening_changes.setdefault(_id, dict())
                opening_changes[_id][timestep] = opening_time
        for e in self.memory.files.get('ProductionResource',[]):
            change = opening_changes.get(e['id'])
            if change:
                e['OpeningTimes'] = change



import dateutil.parser
import os
from typing import Union
import pandas as pd

from azure.kusto.data import KustoClient, KustoConnectionStringBuilder
from azure.kusto.data.data_format import DataFormat
from azure.kusto.ingest import QueuedIngestClient, IngestionProperties


class ADXWrapper:
    """
    Wrapping class to ADX
    """

    @staticmethod
    def type_mapping(key: str, key_example_value) -> str:
        """
        This method is used to replace the type name from python to the one used in ADX
        :param key: the name of the key
        :param key_example_value: a possible value of the key
        :return: the name of the type used in ADX
        """

        if key == "SimulationRun":
            return "guid"

        try:
            # Use dateutil parser to test if the value could be a date, in case of error it is not
            dateutil.parser.parse(key_example_value, fuzzy=False)
            return "datetime"
        except (ValueError, TypeError):
            pass

        if type(key_example_value) is float:
            return "real"

        if type(key_example_value) is int:
            return "long"

        # Default case to string
        return "string"

    def send_to_adx(self, dict_list: list, table_name: str, ignore_table_creation: bool = True, drop_by_tag : str = None) -> bool:
        """
        Will take a list of dict items and send them to a given table in ADX
        :param dict_list: list of dict objects requiring to have the same keys
        :param table_name: The name of the table in which the data should be sent
        :param ignore_table_creation: If set to True won't try to create a table to send the data
        :return: A boolean check if the data have been sent to ADX
        """

        if not ignore_table_creation and not self.table_exists(table_name):
            # If the target table does not exist create it
            # First create the columns types needed for the table
            types = {k: self.type_mapping(k, dict_list[0][k]) for k in dict_list[0].keys()}
            # Then try to create the table
            if not self.create_table(table_name, types):
                print(f"Error creating table {table_name}.")
                return False

        # Create a dataframe with the data to write and send them to ADX
        df = pd.DataFrame(dict_list)
        self.ingest_dataframe(table_name, df, drop_by_tag)
        return True

    def ingest_dataframe(self, table_name: str, dataframe, drop_by_tag : str = None):
        """
        Write the content of dataframe to a table
        :param table_name: name of the target table
        :param dataframe: dataframe containing the data to be written
        :return: None
        """
        drop_by_tags = [drop_by_tag] if (drop_by_tag is not None) else None
        properties = IngestionProperties(database=self.database, table=table_name, data_format=DataFormat.CSV, drop_by_tags=drop_by_tags )
        client = self.ingest_client
        client.ingest_from_dataframe(dataframe, ingestion_properties=properties)

    def run_command_query(self, query: str):
        """
        Execute a command query on the database
        :param query: the query to execute
        :return: the results of the query
        """
        client = self.kusto_client
        return client.execute_mgmt(self.database, query)

    def run_query(self, query: str):
        """
        Execute a simple query on the database
        :param query: the query to execute
        :return: the results of the query
        """
        client = self.kusto_client
        return client.execute(self.database, query)

    def table_exists(self, table_name: str) -> bool:
        """
        Check if a table exists on the database
        :param table_name: The table to look for
        :return: does the table exits ?
        """
        get_tables_query = f".show database ['{self.database}'] schema| distinct TableName"
        tables = self.run_query(get_tables_query)
        for r in tables.primary_results[0]:
            if table_name == r[0]:
                return True
        return False

    def create_table(self, table_name: str, schema: dict) -> bool:
        """
        Create a table on the database
        :param table_name: the name of the table
        :param schema: the schema associated to the table
        :return: Is the table created ?
        """
        # If the table exists no need to create it
        if self.table_exists(table_name):
            return True

        create_query = f".create table {table_name}("
        for column_name, column_type in schema.items():
            create_query += f"{column_name}:{column_type},"
        create_query = create_query[:-1] + ")"
        try:
            self.run_query(create_query)
        except Exception as e:
            print(e)
            return False
        return True

    def __init__(self,
                 database: str,
                 cluster_url: Union[str, None] = None,
                 ingest_url: Union[str, None] = None,
                 cluster_name: Union[str, None] = None,
                 cluster_region: Union[str, None] = None):

        if cluster_name and cluster_region:
            cluster_url = f"https://{cluster_name}.{cluster_region}.kusto.windows.net"
            ingest_url = f"https://ingest-{cluster_name}.{cluster_region}.kusto.windows.net"

        az_client_id = os.environ['AZURE_CLIENT_ID']
        az_client_secret = os.environ['AZURE_CLIENT_SECRET']
        az_tenant_id = os.environ['AZURE_TENANT_ID']

        self.cluster_kcsb = KustoConnectionStringBuilder.with_aad_application_key_authentication(cluster_url,
                                                                                                 az_client_id,
                                                                                                 az_client_secret,
                                                                                                 az_tenant_id)
        self.ingest_kcsb = KustoConnectionStringBuilder.with_aad_application_key_authentication(ingest_url,
                                                                                                az_client_id,
                                                                                                az_client_secret,
                                                                                                az_tenant_id)
        self.kusto_client = KustoClient(self.cluster_kcsb)
        self.ingest_client = QueuedIngestClient(self.ingest_kcsb)
        self.database = database



import json
from Supplychain.Generic.folder_io import FolderWriter
from typing import Union


class JSONWriter(FolderWriter):

    def write_from_list(self, dict_list: list, file_name: str, ordering_key: Union[str, None] = None):
        with open(f"{self.output_folder}/{file_name}.json", "w") as target_file:
            if dict_list:
                to_be_writen = dict_list
                if ordering_key is not None and ordering_key in dict_list[0]:
                    to_be_writen = sorted(dict_list, key=lambda e: e[ordering_key])
                json.dump(list(to_be_writen), target_file, default=str)

    def __init__(self,
                 output_folder: str = "Output"):
        FolderWriter.__init__(self,
                              output_folder=output_folder)



from abc import ABCMeta, abstractmethod


class AbstractOptimizationAlgorithm(object):
    """
    Abstract class allowing the encapsulation of an optimization algorithm
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def generate_decision_variables(self):
        pass

    @abstractmethod
    def update_algorithm(self, dv_list, result_list):
        pass

    @abstractmethod
    def is_not_finished(self):
        pass

    @abstractmethod
    def generate_optimal_solution(self):
        pass


class AbstractObjectiveFunction(object):
    """
    Abstract class allowing the computation of an Objective function using the output of a simulation
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def applies(self, results):
        pass


class AbstractParameterTransformation(object):
    """
    Abstract class allowing the transformation of an Optimization Algorithm output to an input for the simulation
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def applies(self, variables):
        pass


class AbstractOptimization(object):

    __metaclass__ = ABCMeta

    def __init__(self,
                 optimization_algorithm,
                 objective_function,
                 parameter_transformation,
                 simulation_id):

        if not isinstance(optimization_algorithm, AbstractOptimizationAlgorithm):
            raise ValueError("optimization_algorithm must be an OptimizationAlgorithm")

        if not isinstance(objective_function, AbstractObjectiveFunction):
            raise ValueError("objective_function must be an ObjectiveFunction")

        if not isinstance(parameter_transformation, AbstractParameterTransformation):
            raise ValueError("parameter_transformation must be an ParameterTransformation")

        self.opt_algo = optimization_algorithm
        self.obj_func = objective_function
        self.par_tran = parameter_transformation

        self.simulation_id = simulation_id

    @abstractmethod
    def optimize_parameters(self):
        pass



import os
import json
from typing import Union


class FolderWriter:

    @staticmethod
    def json_value(element):
        """
        Dump an element as json for types with a different representation between json and python
        bool : no upper case starting letter
        list : can contain other elements so need to be dumped
        dict : python repr use simple quotes (') instead of double (") as used in json
        :param element: element to be converted
        :return: converted value of the element
        """
        if type(element) in (bool, dict, list):
            return json.dumps(element)
        return element

    def write_from_list(self, dict_list: list, file_name: str, ordering_key: Union[str, None] = None):
        pass

    def __init__(self,
                 output_folder: Union[str, None] = None):

        self.output_folder = output_folder
        if self.output_folder is not None:
            os.makedirs(self.output_folder, exist_ok=True)


class FolderReader:

    def __init__(self,
                 input_folder: str = "Output",
                 keep_nones: bool = True):
        self.input_folder = input_folder

        self.keep_nones = keep_nones

        self.files = None



import csv
import numpy as np
from typing import Union

from Supplychain.Wrappers.simulator import CosmoEngine
from Supplychain.Protocol.cmaes_optimization_algorithm import CMAESOptimization
from Supplychain.Protocol.default_transformation import DefaultTransformation
from Supplychain.Protocol.multiprocessing_optimization import MultiprocessingOptimization
from Supplychain.Protocol.objective_functions import DefaultObjectiveFunction, ProfitMaximizationObjectiveFunction

from Supplychain.Generic.timer import Timer

import json


def get_optimization_objective(input_folder: str = "Input"):
    default_value = "ServiceLevelMaximization"
    with open(input_folder + "/Configuration.csv", "r") as file:
        reader = csv.DictReader(file)
        for line in reader:
            return line.get("OptimizationObjective", default_value)
    return default_value


def run_cmaes_optimization(simulation_name: str,
                           sigma0: float,
                           tol_x: float,
                           tol_fun: float,
                           max_f_evals: int,
                           pop_size: int,
                           amqp_consumer_adress: Union[str, None] = None,
                           optimization_objective: str = "ServiceLevelMaximization"):
    with Timer("[Optimization Runner]") as t:
        simulator = CosmoEngine.LoadSimulator('Simulation')
        time_prefix = "{time_since_last_split:6.4f} ({time_since_start:6.4f}) - "
        t.split(time_prefix + "Load Simulator")

        # Reduce log level to Error during optimization
        logger = CosmoEngine.LoggerManager_GetInstance().GetLogger()
        logger.SetLogLevel(logger.eAssertion)

        # Get data from the simulator
        df_decision_vars = simulator.decision_variables_df
        _end_cycle = simulator.GetModel().GetParameter('NumberOfCycle').GetAsInt()

        nb_pars = 0

        if simulator.FindAttribute("{Model}Model::{Attribute}ActivateVariableMachineOpeningRate").GetAsBool():
            machines = list(sorted(simulator.machine_list))
            nb_pars += len(machines) * simulator.FindAttribute("{Model}Model::{Attribute}NumberOfCycle").GetAsInt()
        else:
            machines = list()

        for group in set(df_decision_vars['Group']):
            nb_var = df_decision_vars[df_decision_vars['Group'] == group].shape[0]
            nb_cycles = (df_decision_vars[df_decision_vars['Group'] == group].reset_index()['EndCycle'][0]
                         - df_decision_vars[df_decision_vars['Group'] == group].reset_index()['StartCycle'][0]
                         + 1)
            nb_pars += nb_cycles * (nb_var if nb_var > 2 else 1)

        nb_pars = max(2, nb_pars)

        specs = dict()
        specs['x0'] = [np.random.random() for _ in range(nb_pars)]
        specs['sigma0'] = sigma0
        specs['inopts'] = dict()
        specs['inopts']['tolx'] = tol_x
        specs['inopts']['tolfun'] = tol_fun
        specs['inopts']['maxfevals'] = max_f_evals
        specs['inopts']['popsize'] = pop_size

        # Initialization of 'bricks'
        # Initialization of CMAES: the optimization alogrithm used
        algorithm = CMAESOptimization(parameter_count=nb_pars,
                                      **specs)
        t.split(time_prefix + "Prepare CMAES")

        # Initialization of the transformation which takes CMAES results and transform them on map: datapath -> value
        transformation = DefaultTransformation(df_decision_vars,
                                               machines,
                                               _end_cycle)
        t.split(time_prefix + "Prepare transformation")

        objective_function = DefaultObjectiveFunction()

        if optimization_objective == "ProfitMaximization":
            objective_function = ProfitMaximizationObjectiveFunction()
        t.split(time_prefix + "Prepare Objective Function")

        optim = MultiprocessingOptimization(algorithm,
                                            objective_function,
                                            transformation,
                                            simulation_name,
                                            pop_size,
                                            amqp_consumer_adress)
        t.split(time_prefix + "Prepare Optimization")

        optim.optimize_parameters()

        # Get optimized results for the algorithm after optimization
        optim_dvs = transformation.applies(algorithm.generate_optimal_solution(), 0)
        t.split(time_prefix + "Optimal solution found")

        # Applies optimized results to the simulator
        for datapath, stringvalue in optim_dvs.items():
            simulator.FindAttribute(datapath).SetAsString(stringvalue)

        if amqp_consumer_adress is not None:
            # Remove old consumers
            for consumer in simulator.GetConsumers():
                simulator.DestroyConsumer(consumer)

            # Instantiate consumers using AMQP to send data to the cloud service
            simulator.InstantiateAMQPConsumers(simulation_name, amqp_consumer_adress)

        # Put back log level to Info for final simulation
        logger.SetLogLevel(logger.eInfo)
        t.split(time_prefix + "Optimal simulation ready")

        # Run final simulation
        simulator.Run()
        t.split(time_prefix + "Optimal simulation done")



parameters_default_values = {
    'Stock': {
        #'id',
        #'Label',
        'PlantName': '',
        'Step': '',
        'PartId': '',
        'MinimalStock': -1,
        'MaximalStock': None,
        #'InitialStock',
        #'InitialValue',
        #'IsInfinite',
        'PurchasingUnitCost': 0,
        'CO2UnitEmissions': 0,
        'UnitIncome': 0,
        'BacklogWeight': 1,
        'MaximizationWeight': 0,
        'StockPolicy': 'None',
        'SourcingPolicy': 'Equidistribution',
        'DispatchPolicy': 'Equidistribution',
        'ReviewPeriod': 1,
        'NextReview': 0,
        'Advance': 0,
        'StorageUnitCosts': {},
        'OrderPoints': {},
        'OrderQuantities': {},
        'OrderUpToLevels': {},
        'SafetyQuantities': {},
        'Demands': {},
        'DemandUncertainties': {},
        'DemandWeights': {},
        'SalesForecasts': {},
    },
    'Transport': {
        #'Label',
        #'source',
        #'target',
        #'Duration',
        'Priority': 0,
        'CustomFees': {},
        'TransportUnitCosts': {},
        'CO2UnitEmissions': {},
        'InitialTransportedQuantities': {},
        'InitialTransportedValues': {},
    },
    'Configuration': {
        'ActivateUncertainties': False,
        'ActivateVariableMachineOpeningRate': False,
        'BatchSize': 0,
        'EmptyObsoleteStocks': False,
        'EnforceProductionPlan': False,
        'FinancialCostOfStock': 0,
        'ManageBacklogQuantities': False,
        'OptimizationObjective': 'ServiceLevelMaximization',
        'SimulatedCycles': 1,
        #'StartingDate',
        'StepsPerCycle': 1,
        'TimeStepDuration': 1440,
        'UncertaintiesProbabilityDistribution': 'Uniform',
        'IntermediaryStockDispatchPolicy': 'DispatchAll',
    },
    'contains': {
        #'source',
        #'target',
    },
    'output': {
        #'source',
        #'target',
    },
    'input': {
        #'source',
        #'target',
        'InputQuantity': 1,
    },
    'ProductionResource': {
        #'id',
        #'Label',
        'PlantName': '',
        'ProductionStep': '',
        'ProductionPolicy': 'Equidistribution',
        'FixedProductionCosts': {},
        #'OpeningTimes',
    },
    'ProductionOperation': {
        #'id',
        #'Label',
        'PlantName': '',
        'IsContractor': False,
        'InvestmentCost': 0,
        'Priority': 0,
        'Duration': 0,
        'QuantitiesToProduce': {},
        'OperatingPerformances': {},
        'CycleTimes': {},
        'RejectRates': {},
        'OperatingPerformanceUncertainties': {},
        'ProductionUnitCosts': {},
        'CO2UnitEmissions': {},
    }
}
variables_default_values = {
    'Stock': {
        'StorageUnitCosts': 0,
        'OrderPoints': 0,
        'OrderQuantities': 0,
        'OrderUpToLevels': 0,
        'SafetyQuantities': 0,
        'Demands': 0,
        'DemandUncertainties': 0,
        'DemandWeights': 1,
        'SalesForecasts': 0,
    },
    'Transport': {
        'CustomFees': 0,
        'TransportUnitCosts': 0,
        'CO2UnitEmissions': 0,
        'InitialTransportedQuantities': 0,
        'InitialTransportedValues': 0,
    },
    'Configuration': {
    },
    'contains': {
    },
    'output': {
    },
    'input': {
    },
    'ProductionResource': {
        'FixedProductionCosts': 0,
    },
    'ProductionOperation': {
        'QuantitiesToProduce': 0,
        'OperatingPerformances': 1,
        'CycleTimes': 1,
        'RejectRates': 0,
        'OperatingPerformanceUncertainties': 0,
        'ProductionUnitCosts': 0,
        'CO2UnitEmissions': 0,
    }
}



class ValidationSchemas:
    schemas = {
        "Configuration": {
            "$schema": "http://json-schema.org/draft-07/schema#",
            "type": "object",
            "properties": {
                "StartingDate": {
                    "type": "string",
                    "format": "date"
                },
                "SimulatedCycles": {
                    "type": "integer",
                    "minimum": 1
                },
                "StepsPerCycle": {
                    "type": "integer",
                    "minimum": 1
                },
                "TimeStepDuration": {
                    "type": "integer",
                    "minimum": 1
                },
                "ManageBacklogQuantities": {
                    "type": "boolean"
                },
                "OptimizationObjective": {
                    "type": "string",
                    "enum": ["ServiceLevelMaximization",
                             "ProfitMaximization"]
                },
                "ActivateUncertainties": {
                    "type": "boolean"
                },
                "UncertaintiesProbabilityDistribution": {
                    "type": "string",
                    "enum": ["Uniform",
                             "Gaussian"]
                },
                "FinancialCostOfStock": {
                    "type": "number",
                    "minimum": 0,
                    "maximum": 1
                },
                "BatchSize": {
                    "type": "integer",
                    "minimum": 0
                },
                "EmptyObsoleteStocks": {
                    "type": "boolean"
                },
                "ActivateVariableMachineOpeningRate": {
                    "type": "boolean"
                },
                "EnforceProductionPlan": {
                    "type": "boolean"
                },
                "IntermediaryStockDispatchPolicy": {
                    "type": "string",
                    "enum": ["DispatchAll",
                             "AllowRetention"]
                },
            }
        },
        "input": {
            "$schema": "http://json-schema.org/draft-07/schema#",
            "type": "object",
            "properties": {
                "InputQuantity": {
                    "type": "number",
                    "exclusiveMinimum": 0
                }
            }
        },
        "ProductionOperation": {
            "$schema": "http://json-schema.org/draft-07/schema#",
            "type": "object",
            "properties": {
                "QuantitiesToProduce": {
                    "type": "object",
                    "patternProperties": {
                        "^[0-9]+$": {
                            "type": "number",
                            "minimum": 0
                        }
                    },
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]}
                },
                "OperatingPerformances": {
                    "type": "object",
                    "patternProperties": {
                        "^[0-9]+$": {
                            "type": "number",
                            "minimum": 0,
                            "maximum": 1
                        }
                    },
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]}

                },
                "CycleTimes": {
                    "type": "object",
                    "patternProperties": {
                        "^[0-9]+$": {
                            "type": "number",
                            "minimum": 0
                        }
                    },
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]}
                },
                "RejectRates": {
                    "type": "object",
                    "patternProperties": {
                        "^[0-9]+$": {
                            "type": "number",
                            "minimum": 0,
                            "maximum": 1
                        }
                    },
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]}
                },
                "OperatingPerformanceUncertainties": {
                    "type": "object",
                    "patternProperties": {
                        "^[0-9]+$": {
                            "type": "number",
                            "minimum": 0,
                            "maximum": 1
                        }
                    },
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]}
                },
                "ProductionUnitCosts": {
                    "type": "object",
                    "patternProperties": {
                        "^[0-9]+$": {
                            "type": "number",
                            "minimum": 0
                        }
                    },
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]}
                },
                "CO2UnitEmissions": {
                    "type": "object",
                    "patternProperties": {
                        "^[0-9]+$": {
                            "type": "number",
                            "minimum": 0
                        }
                    },
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]}
                },
                "IsContractor": {
                    "type": "boolean"
                },
                "InvestmentCost": {
                    "type": "number",
                    "minimum": 0,
                },
                "Priority": {
                    "type": "integer",
                    "minimum": 0
                },
                "Duration": {
                    "type": "integer",
                    "minimum": 0
                },
            },
            "required": ["CycleTimes"],
            "dependencies": {
                "OperatingPerformanceUncertainties": ["OperatingPerformances"]
            }
        },
        "ProductionResource": {
            "$schema": "http://json-schema.org/draft-07/schema#",
            "type": "object",
            "properties": {
                "ProductionPolicy": {
                    "type": "string",
                    "enum": ["None",
                             "Equidistribution",
                             "GreatestWorkload",
                             "SmallestWorkload",
                             "HighestPriority"]
                },
                "OpeningTimes": {
                    "type": "object",
                    "patternProperties": {
                        "^[0-9]+$": {
                            "type": "number",
                            "minimum": 0
                        }
                    },
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]}
                },
                "FixedProductionCosts": {
                    "type": "object",
                    "patternProperties": {
                        "^[0-9]+$": {
                            "type": "number",
                            "minimum": 0
                        }
                    },
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]}
                }
            },
            "required": ["OpeningTimes"]
        },
        "Stock": {
            "$schema": "http://json-schema.org/draft-07/schema#",
            "type": "object",
            "properties": {
                "StorageCosts": {
                    "type": "object",
                    "patternProperties": {
                        "^[0-9]+$": {
                            "type": "number",
                            "minimum": 0
                        }
                    },
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]}
                },
                "IsInfinite": {
                    "type": "boolean"
                },
                "MinimalStock": {
                    "type": "number",
                    "if": {"maximum": -1},
                    "then": {"minimum": -1},
                    "else": {"minimum": 0}
                },
                "MaximalStock": {
                    "type": "number",
                    "if": {"maximum": -1},
                    "then": {"minimum": -1},
                    "else": {"minimum": 0}
                },
                "InitialStock": {
                    "type": "number",
                    "minimum": 0
                },
                "InitialValue": {
                    "type": "number",
                    "minimum": 0
                },
                "PurchasingUnitCost": {
                    "type": "number",
                    "minimum": 0
                },
                "CO2UnitEmissions": {
                    "type": "number",
                    "minimum": 0
                },
                "UnitIncome": {
                    "type": "number",
                    "minimum": 0
                },
                "Demands": {
                    "type": "object",
                    "patternProperties": {
                        "^[0-9]+$": {
                            "type": "number",
                            "minimum": 0
                        }
                    }
                },
                "DemandUncertainties": {
                    "type": "object",
                    "patternProperties": {
                        "^[0-9]+$": {
                            "type": "number",
                            "minimum": 0,
                            "maximum": 1
                        }
                    }
                },
                "DemandWeights": {
                    "type": "object",
                    "patternProperties": {
                        "^[0-9]+$": {
                            "type": "number",
                            "minimum": 0
                        }
                    }
                },
                "BacklogWeight": {
                    "type": "number",
                    "minimum": 0
                },
                "MaximizationWeight": {
                    "type": "number",
                    "minimum": 0
                },
                "StockPolicy": {
                    "type": "string",
                    "enum": ["None",
                             "OrderPointFixedQuantity",
                             "OrderPointVariableQuantity",
                             "MakeToForecast"]
                },
                "SourcingPolicy": {
                    "type": "string",
                    "enum": ["Equidistribution",
                             "HighestStock",
                             "HighestPriority"]
                },
                "DispatchPolicy": {
                    "type": "string",
                    "enum": ["None",
                             "Equidistribution",
                             "GreatestQuantity",
                             "SmallestQuantity",
                             "HighestPriority"]
                },
                "ReviewPeriod": {
                    "type": "integer",
                    "minimum": 1
                },
                "NextReview": {
                    "type": "integer",
                    "minimum": 0
                },
                "OrderPoints": {
                    "type": "object",
                    "patternProperties": {
                        "^[0-9]+$": {
                            "type": "number",
                            "minimum": 0
                        }
                    },
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]}
                },
                "OrderQuantities": {
                    "type": "object",
                    "patternProperties": {
                        "^[0-9]+$": {
                            "type": "number",
                            "minimum": 0
                        }
                    },
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]}
                },
                "OrderUpToLevels": {
                    "type": "object",
                    "patternProperties": {
                        "^[0-9]+$": {
                            "type": "number",
                            "minimum": 0
                        }
                    },
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]}
                },
                "Advance": {
                    "type": "integer",
                    "minimum": 0
                },
                "SafetyQuantities": {
                    "type": "object",
                    "patternProperties": {
                        "^[0-9]+$": {
                            "type": "number",
                            "minimum": 0
                        }
                    },
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]}
                },
                "SalesForecasts": {
                    "type": "object",
                    "patternProperties": {
                        "^[0-9]+$": {
                            "type": "number",
                            "minimum": 0
                        }
                    }
                },
            },
            "dependencies": {
                "InitialValue": ["InitialStock"]
            }
        },
        "Transport": {
            "$schema": "http://json-schema.org/draft-07/schema#",
            "type": "object",
            "properties": {
                "Duration": {
                    "type": "integer",
                    "minimum": 0
                },
                "Priority": {
                    "type": "integer",
                    "minimum": 0
                },
                "InitialTransportedQuantities": {
                    "type": "object",
                    "patternProperties": {
                        "^[0-9]+$": {
                            "type": "number",
                            "minimum": 0
                        }
                    }
                },
                "InitialTransportedValues": {
                    "type": "object",
                    "patternProperties": {
                        "^[0-9]+$": {
                            "type": "number",
                            "minimum": 0
                        }
                    }
                },
                "CustomFees": {
                    "type": "object",
                    "patternProperties": {
                        "^[0-9]+$": {
                            "type": "number",
                            "minimum": 0
                        }
                    },
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]}
                },
                "TransportUnitCosts": {
                    "type": "object",
                    "patternProperties": {
                        "^[0-9]+$": {
                            "type": "number",
                            "minimum": 0
                        }
                    },
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]}
                },
                "CO2UnitEmissions": {
                    "type": "object",
                    "patternProperties": {
                        "^[0-9]+$": {
                            "type": "number",
                            "minimum": 0
                        }
                    },
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]}
                }
            },
            "required": ["Duration"],
            "dependencies": {
                "InitialTransportedValues": ["InitialTransportedQuantities"]
            }
        }
    }
    graph = {
        "contains": {
            "source": "source",
            "target": "target",
            "links": {
                "source": ["ProductionResource", "id"],
                "target": ["ProductionOperation", "id"]
            },
            "cardinalities": "1:N",
            "all_target_present": True,
            "all_source_present": True
        },
        "input": {
            "source": "source",
            "target": "target",
            "links": {
                "source": ["Stock", "id"],
                "target": ["ProductionOperation", "id"]
            },
            "cardinalities": "N:N",
            "all_target_present": True,
            "all_source_present": False
        },
        "output": {
            "source": "source",
            "target": "target",
            "links": {
                "source": ["ProductionOperation", "id"],
                "target": ["Stock", "id"]
            },
            "cardinalities": "N:1",
            "all_target_present": False,
            "all_source_present": True
        },
        "Transport": {
            "source": "source",
            "target": "target",
            "links": {
                "source": ["Stock", "id"],
                "target": ["Stock", "id"]
            },
            "cardinalities": "N:N",
            "all_target_present": False,
            "all_source_present": False
        }
    }



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
import os
from collections import namedtuple
# environ change to avoid a crash on multiple simulations: PROD-5387
os.environ['CSM_METRICS_MODE'] = "off"

import pandas

import csm.engine as CosmoEngine


def register_into_simulatorinterface(method):
    if hasattr(method, "fget"):
        name = method.fget.__name__
    else:
        name = method.__name__
    setattr(CosmoEngine.SimulatorInterface, name, method)
    return method


@register_into_simulatorinterface
@property
def machine_list(self):
    return [
        e.GetName()
        for e in self.GetModel().FindEntitiesByType("Machine")
    ]


@register_into_simulatorinterface
@property
def _decision_variables(self):
    if self.FindAttribute("{Model}Model::{Attribute}AllowPartRetention").GetAsBool():
        entities_by_name = self.get_entities_by_name()
        (_, pair_to_sto) = self.get_stocks_to_operations_pairs()
        stock_list_sto = dict()
        for pair, sto in pair_to_sto.items():
            stock_list_sto.setdefault(pair[0], []).append(sto.GetName())
        for stock, stos in stock_list_sto.items():
            stock_groups = set()
            for sto_name in stos:
                sto = entities_by_name[sto_name]
                groups = CosmoEngine.DataTypeMapInterface_Cast(sto.GetAttribute("Groups"))
                name = sto.GetName()
                for k in map(str, groups.GetKeys()):
                    yield (
                        "StockToOperation",
                        name,
                        None,
                        "PropStockForOperation",
                        groups.GetAttribute(k).GetAttributeAsString("StartCycle"),
                        groups.GetAttribute(k).GetAttributeAsString("EndCycle"),
                        k
                    )
                    stock_groups.add((k, groups.GetAttribute(k).GetAttributeAsString("StartCycle"), groups.GetAttribute(k).GetAttributeAsString("EndCycle")))
            for key, start_cycle, end_cycle in stock_groups:
                yield (
                    "Stock",
                    stock,
                    None,
                    None,
                    start_cycle,
                    end_cycle,
                    key
                )
    else:
        for sto in self.GetModel().FindEntitiesByType("StockToOperation"):
            groups = CosmoEngine.DataTypeMapInterface_Cast(sto.GetAttribute("Groups"))
            name = sto.GetName()
            for k in map(str, groups.GetKeys()):
                yield (
                    "StockToOperation",
                    name,
                    None,
                    "PropStockForOperation",
                    groups.GetAttribute(k).GetAttributeAsString("StartCycle"),
                    groups.GetAttribute(k).GetAttributeAsString("EndCycle"),
                    k
                )
    for machine in self.GetModel().FindEntitiesByType("Machine"):
        groups = CosmoEngine.DataTypeMapInterface_Cast(machine.GetAttribute("Groups"))
        for operation in CosmoEngine.CompoundEntity_Cast(machine).FindEntitiesByType("Operation"):
            name = machine.GetName()
            op_name = operation.GetName()
            for k in map(str, groups.GetKeys()):
                yield (
                    "Machine",
                    name,
                    op_name,
                    "StockTypeProportions",
                    groups.GetAttribute(k).GetAttributeAsString("StartCycle"),
                    groups.GetAttribute(k).GetAttributeAsString("EndCycle"),
                    k
                )


@register_into_simulatorinterface
@property
def decision_variables_df(self):
    df = pandas.DataFrame(
        self._decision_variables,
        columns=[
            "Type",
            "Entity",
            "Operation",
            "Attribute",
            "StartCycle",
            "EndCycle",
            "Group",
        ]
    )
    df["StartCycle"] = df["StartCycle"].astype(int)
    df["EndCycle"] = df["EndCycle"].astype(int)
    df["Group"] = df["Group"].astype(int)
    return df


@register_into_simulatorinterface
@property
def _constraints(self):
    for stock in self.GetModel().FindEntitiesByType("Stock"):
        demand = CosmoEngine.DataTypeMapInterface_Cast(stock.GetAttribute("Demand"))
        for k in map(str, demand.GetKeys()):
            yield (
                stock.GetName(),
                demand.GetAttribute(k).GetAttributeAsString("ExternalDemand"),
                demand.GetAttribute(k).GetAttributeAsString("InternalDemand"),
                k,
                demand.GetAttribute(k).GetAttributeAsString("ExternalWeight"),
                demand.GetAttribute(k).GetAttributeAsString("InternalWeight"),
                demand.GetAttribute(k).GetAttributeAsString("WeightMax"),
                demand.GetAttribute(k).GetAttributeAsString("BacklogWeight"),
                demand.GetAttribute(k).GetAttributeAsString("MaxVal")
            )


@register_into_simulatorinterface
@property
def constraints_df(self):
    return pandas.DataFrame(self._constraints, columns=[
        "name",
        "demand_extern",
        "demand_intern",
        "cycle",
        "weight_extern",
        "weight_intern",
        "weighto",
        "max_val",
        "weight_backlog",
    ])


setattr(CosmoEngine.SimulatorInterface, '_model_parameters', None)


@register_into_simulatorinterface
def get_model_parameters(self, force=False):
    if force or self._model_parameters is None:
        model = self.GetModel()
        parameters = model.GetParameters()
        self._model_parameters = {
            parameter: (
                parameters.GetAttribute(parameter).Get()
                if parameter_type not in ('Duration', 'DateTime')
                else parameters.GetAttribute(parameter).GetAsString()
            )
            for parameter, parameter_type in model.GetParametersDefinition()
        }
    return self._model_parameters


setattr(CosmoEngine.SimulatorInterface, '_number_of_time_steps', None)


@register_into_simulatorinterface
def get_number_of_time_steps(self, force=False):
    if force or self._number_of_time_steps is None:
        model_parameters = self.get_model_parameters(force=force)
        self._number_of_time_steps = model_parameters['TimeStepPerCycle'] * model_parameters['NumberOfCycle']
    return self._number_of_time_steps


setattr(CosmoEngine.SimulatorInterface, '_entities_by_name', None)


@register_into_simulatorinterface
def get_entities_by_name(self, force=False):
    if force or self._entities_by_name is None:
        self._entities_by_name = {
            entity.GetName(): entity
            for entity in self.GetModel().GetSubEntities()
        }
    return self._entities_by_name


setattr(CosmoEngine.SimulatorInterface, '_entities_by_type', None)


@register_into_simulatorinterface
def get_entities_by_type(self, entity_type=None, force=False):
    if force or self._entities_by_type is None:
        model = self.GetModel()
        self._entities_by_type = {
            entity_type: []
            for entity_type in model.GetEntityTypeList()
        }
        for entity in model.GetSubEntities():
            self._entities_by_type[entity.GetTypeName()].append(entity)
    return (
        self._entities_by_type
        if entity_type is None
        else self._entities_by_type[entity_type]
    )


@register_into_simulatorinterface
def get_typed_entities_by_name(self, entity_type, force=False):
    return {
        entity.GetName(): entity
        for entity in self.get_entities_by_type(entity_type=entity_type, force=force)
    }


setattr(CosmoEngine.SimulatorInterface, '_entities_names_by_type', None)


@register_into_simulatorinterface
def get_entities_names_by_type(self, entity_type=None, force=False):
    if force or self._entities_names_by_type is None:
        self._entities_names_by_type = {
            entities_type: {
                entity.GetName()
                for entity in entities
            }
            for entities_type, entities in self.get_entities_by_type(force=force).items()
        }
    return (
        self._entities_names_by_type
        if entity_type is None
        else self._entities_names_by_type[entity_type]
    )


setattr(CosmoEngine.SimulatorInterface, '_pairs_by_stock_to_operation', None)
setattr(CosmoEngine.SimulatorInterface, '_stocks_to_operations_by_pair', None)


@register_into_simulatorinterface
def get_stocks_to_operations_pairs(self, force=False):
    if force or self._pairs_by_stock_to_operation is None or self._stocks_to_operations_by_pair is None:
        self._pairs_by_stock_to_operation = {}
        self._stocks_to_operations_by_pair = {}
        for stock_to_operation in self.get_entities_by_type('StockToOperation', force=force):
            edge = CosmoEngine.Edge_Cast(stock_to_operation)
            stock = edge.GetLeft()
            operation = edge.GetRight()
            self._pairs_by_stock_to_operation[stock_to_operation.GetName()] = (stock, operation)
            self._stocks_to_operations_by_pair[(stock.GetName(), operation.GetName())] = stock_to_operation
    return self._pairs_by_stock_to_operation, self._stocks_to_operations_by_pair


setattr(CosmoEngine.SimulatorInterface, '_entities_neighbors_by_type', None)


@register_into_simulatorinterface
def get_entities_neighbors_by_type(self, entity_type=None, force=False):
    if force or self._entities_neighbors_by_type is None:
        industrial_network, = self.get_entities_by_type('IndustrialNetwork', force=force)
        environment = self.GetModel().GetCompoundEntity(industrial_network.GetId()).GetEnvironment()
        self._entities_neighbors_by_type = {
            entities_type: {}
            for entities_type in ('Stock', 'Operation', 'TransportOperation')
        }
        for entities_type, typed_entities_neighbors in self._entities_neighbors_by_type.items():
            for entity in self.get_entities_by_type(entities_type, force=force):
                if not environment.GetNeighborsCount(entity):
                    continue
                entity_neighbors = typed_entities_neighbors.setdefault(entity.GetName(), {})
                for neighbor in environment.GetNeighbors(entity):
                    entity_neighbors.setdefault(neighbor.GetTypeName(), set()).add(neighbor.GetName())
    return (
        self._entities_neighbors_by_type
        if entity_type is None
        else self._entities_neighbors_by_type[entity_type]
    )


setattr(CosmoEngine.SimulatorInterface, '_stocks_by_operation', None)


@register_into_simulatorinterface
def get_stocks_by_operation(self, force=False):
    if force or self._stocks_by_operation is None:
        self._stocks_by_operation = {}
        entities_neighbors_by_type = self.get_entities_neighbors_by_type(force=force)
        stocks_neighbors = entities_neighbors_by_type['Stock']
        operations_neighbors = entities_neighbors_by_type['Operation'].copy()
        operations_neighbors.update(entities_neighbors_by_type['TransportOperation'])
        for operation_name, operation_neighbors in operations_neighbors.items():
            input_stocks = operation_neighbors['Stock']
            output_stocks = set(
                stock_name
                for stock_name, stock_neighbors in stocks_neighbors.items()
                if any(
                    neighbor_operation_name == operation_name
                    for entity_type in ('Operation', 'TransportOperation')
                    for neighbor_operation_name in stock_neighbors.get(entity_type, set())
                )
            )
            output_stocks -= input_stocks
            self._stocks_by_operation[operation_name] = (input_stocks, output_stocks)
    return self._stocks_by_operation


setattr(CosmoEngine.SimulatorInterface, '_typed_operations_by_output_stock', None)


@register_into_simulatorinterface
def get_typed_operations_by_output_stock(self, force=False):
    if force or self._typed_operations_by_output_stock is None:
        entities_neighbors_by_type = self.get_entities_neighbors_by_type(force=force)
        stocks_neighbors = {
            stock_name: stock_neighbors.get('Operation', set()) | stock_neighbors.get('TransportOperation', set())
            for stock_name, stock_neighbors in entities_neighbors_by_type['Stock'].items()
        }
        operations_neighbors = entities_neighbors_by_type['Operation'].copy()
        operations_neighbors.update(entities_neighbors_by_type['TransportOperation'])
        input_stocks_by_operation = {
            operation_name: operation_neighbors['Stock']
            for operation_name, operation_neighbors in operations_neighbors.items()
        }
        self._typed_operations_by_output_stock = {}
        entities_by_name = self.get_entities_by_name(force=force)
        for stock_name, stock_neighbors in stocks_neighbors.items():
            upstream_operations_names = [
                operation_name
                for operation_name in stock_neighbors
                if stock_name not in input_stocks_by_operation[operation_name]
            ]
            if upstream_operations_names:
                output_stock_typed_operations = self._typed_operations_by_output_stock.setdefault(stock_name, {})
                for operation_name in upstream_operations_names:
                    output_stock_typed_operations.setdefault(entities_by_name[operation_name].GetTypeName(), set()).add(operation_name)
    return self._typed_operations_by_output_stock


setattr(CosmoEngine.SimulatorInterface, '_transports_by_stocks', None)


@register_into_simulatorinterface
def get_transports_by_stocks(self, force=False):
    if force or self._transports_by_stocks is None:
        self._transports_by_stocks = {}
        entities_by_name = self.get_entities_by_name(force=force)
        stocks_neighbors = self.get_entities_neighbors_by_type('Stock', force=force)
        for transport_name, transport_neighbors in self.get_entities_neighbors_by_type('TransportOperation', force=force).items():
            input_stock_name, = transport_neighbors['Stock']
            for output_stock_name, output_stock_neighbors in stocks_neighbors.items():
                if output_stock_name != input_stock_name and transport_name in output_stock_neighbors.get('TransportOperation', set()):
                    break
            self._transports_by_stocks[input_stock_name, output_stock_name] = entities_by_name[transport_name]
    return self._transports_by_stocks


setattr(CosmoEngine.SimulatorInterface, '_machines_by_operation', None)


@register_into_simulatorinterface
def get_machines_by_operation(self, force=False):
    if force or self._machines_by_operation is None:
        self._machines_by_operation = {
            operation.GetName(): machine
            for machine in self.get_entities_by_type('Machine', force=force)
            for operation in self.GetModel().GetCompoundEntity(machine.GetId()).GetSubEntities()
        }
    return self._machines_by_operation


setattr(CosmoEngine.SimulatorInterface, '_operations_by_machine', None)


@register_into_simulatorinterface
def get_operations_by_machine(self, force=False):
    if force or self._operations_by_machine is None:
        self._operations_by_machine = {
            machine.GetName(): [operation for operation in self.GetModel().GetCompoundEntity(machine.GetId()).GetSubEntities()]
            for machine in self.get_entities_by_type('Machine', force=force)
        }
    return self._operations_by_machine


@register_into_simulatorinterface
def get_scheduled_values(self, entity, schedulable_parameter_name, force=False):
    number_of_time_steps = self.get_number_of_time_steps(force=force)
    default_value = entity.GetAttribute(schedulable_parameter_name).Get()
    scheduled_values = [default_value] * number_of_time_steps
    changes = CosmoEngine.DataTypeMapInterface_Cast(entity.GetAttribute(f'{schedulable_parameter_name}Schedule')).Get()
    if changes:
        time_steps = sorted(time_step for time_step in changes)
        next_time_steps = [*time_steps[1:], number_of_time_steps]
        for t1, t2 in zip(time_steps, next_time_steps):
            scheduled_values[t1:t2] = [changes[t1]] * (t2 - t1)
    return scheduled_values


setattr(CosmoEngine.SimulatorInterface, '_schedulable_parameters_names', None)


@register_into_simulatorinterface
def get_schedulable_parameters_names(self, force=False):
    if force or self._schedulable_parameters_names is None:
        self._schedulable_parameters_names = {}
        suffix = 'Schedule'
        suffix_length = len(suffix)
        for entity_type, entities in self.get_entities_by_type(force=force).items():
            entity = entities[0]
            attributes_names = set(attribute_name for attribute_name, _ in entity.GetStateDefinition())
            entity_schedulable_parameters = [
                schedulable_parameter_name
                for schedulable_parameter_name in (
                    attribute_name[:-suffix_length]
                    for attribute_name in attributes_names
                    if attribute_name.endswith(suffix)
                )
                if schedulable_parameter_name in attributes_names
            ]
            if entity_schedulable_parameters:
                self._schedulable_parameters_names[entity_type] = entity_schedulable_parameters
    return self._schedulable_parameters_names


Demand = namedtuple('Demand', ['stock', 'time_step', 'external_demand', 'internal_demand', 'maximum_quantity', 'external_weight', 'internal_weight', 'backlog_weight', 'maximization_weight'])


@register_into_simulatorinterface
def get_demands(self, force=False):
    for stock in self.get_entities_by_type('Stock', force=force):
        demands = CosmoEngine.DataTypeMapInterface_Cast(stock.GetAttribute("Demand"))
        stock_name = stock.GetName()
        for time_step in demands.GetKeys():
            demand = demands.GetAt(time_step)
            yield Demand(
                stock_name,
                time_step.Get(),
                demand.GetAttribute("ExternalDemand").Get(),
                demand.GetAttribute("InternalDemand").Get(),
                demand.GetAttribute("MaxVal").Get(),
                demand.GetAttribute("ExternalWeight").Get(),
                demand.GetAttribute("InternalWeight").Get(),
                demand.GetAttribute("BacklogWeight").Get(),
                demand.GetAttribute("WeightMax").Get(),
            )



from collections import defaultdict

import numpy as np
import pandas as pd

from Supplychain.Protocol.protocol import AbstractParameterTransformation


class DefaultTransformation(AbstractParameterTransformation):
    """
    A class used to generate and store the transformation applied to the generic algorithm output

    We need a list of decision variables which allows us to transform an array of float to a map of datapath to values
    """

    class BasicTransform(object):
        """
        Define a basic transformation of the data using :
        - a starting index
        - a size of the data to use
        - a list of entities
        - an index in the final array
        """
        index = None
        size = None
        entities = None
        end_index = None

        def __init__(self, index, size, entities, end_index=0):
            self.index = index
            self.size = size
            self.entities = entities
            self.end_index = end_index

        def transform(self, d):
            """

            :param d: the entrance datas
            :return: the transformed datas
            """
            pass

        def __str__(self):
            return str(self.index) + " - " + str(self.size) + " - " + str(self.end_index) + " : " + str(self.entities)

    class InverseTransform(BasicTransform):
        """
        A specialisation of BasicTransform used to generate inverted data
        """

        def transform(self, data):
            """
            For the first entity we keep the original data and for the second we return the inverse of the data
            :param data: an array of data
            :return: an array of tuple : entity, transformed data
            """
            return [(self.entities[0], data[self.index: self.index + self.size]),
                    (self.entities[1], 1 - data[self.index: self.index + self.size])]

    class NormalisationTransform(BasicTransform):
        """
        A specialisation of BasicTransform used to normalize data
        """

        def transform(self, data):
            """
            For each cycle we normalize the data before distributing it to each entity
            :param data: an array of data
            :return: an array of tuple : entity, transformed data
            """
            n_entities = len(self.entities)
            tmp = data[self.index: self.index + self.size * n_entities].reshape((self.size, n_entities))
            tmp = list((tmp / np.linalg.norm(tmp, axis=1, ord=1)[:, None]).T)
            return zip(self.entities, tmp)

    class NoTransform(BasicTransform):
        """
        A specialisation of Basic transform that map the data for the entity without transformations
        """

        def transform(self, data):
            return [(self.entities[0], data[self.index: self.index + self.size]), ]

    transformation_template = None
    start_cycle = None
    array_size = None

    @staticmethod
    def rowToKey(row):
        """
        :param row: an entry of the decision_variable.csv
        :return: a tuple (datapath , <optional> subentity)
        """
        entity_name = row['Entity']
        attribute_name = row['Attribute']
        if attribute_name is None:
            return None
        key = 'Model::{Entity}IndustrialNetwork::{Entity}' + entity_name + '::@' + attribute_name
        if row['Type'] == 'Machine':
            sub_entity_name = row['Operation']
            return key, sub_entity_name
        return key,

    def __init__(self, dv_df, machines, total_cycles):
        """
        We initialize the transformation using the decision variables
        :param dv_df: A pandas DF using decision_variable.csv
        """
        self.transformation_template = list()
        index = 0
        self.start_cycle = 0
        self.array_size = total_cycles

        if not dv_df.empty:
            nb_group = max(list(dv_df['Group'])) + 1
            # Loop on the different groups
            for group in range(nb_group):
                temp = dv_df[dv_df['Group'] == group]
                temp.index = pd.RangeIndex(len(temp.index))

                cycles = temp['EndCycle'][0] - temp['StartCycle'][0] + 1
                final_index = temp['StartCycle'][0] - self.start_cycle
                entries = []
                for _, row in temp.iterrows():
                    entries.append(self.rowToKey(row))

                nb_entries = len(entries)
                if nb_entries == 2:
                    self.transformation_template.append(self.InverseTransform(index=index,
                                                                              size=cycles,
                                                                              entities=entries,
                                                                              end_index=final_index))
                    index += cycles
                else:
                    self.transformation_template.append(self.NormalisationTransform(index=index,
                                                                                    size=cycles,
                                                                                    entities=entries,
                                                                                    end_index=final_index))
                    index += cycles * nb_entries
        for machine in machines:
            machine_data_path = ("Model::{Entity}IndustrialNetwork::{Entity}" + machine + "::@OpeningRates",)
            self.transformation_template.append(self.NoTransform(index=index,
                                                                 size=total_cycles,
                                                                 entities=[machine_data_path, ],
                                                                 end_index=0))
            index += total_cycles

    def applies(self, _data, _seed=0):
        """
        We apply the tranformation on the data
        :param _data: an array of data
        :param _seed: an int used as the seed for the simulation
        :return: a dict ready to be given to the simulation
        """
        data = np.array(_data)
        ret_dict = defaultdict(list)
        if _seed > 0:
            ret_dict["{Model}Model::{Attribute}Seed"] = _seed

        for t in self.transformation_template:
            index = t.end_index
            padding = t.size
            for a, b in t.transform(data):
                if a is None:
                    # Non used DVs have entity names set to None
                    continue
                if len(a) == 1:
                    if not len(ret_dict[a[0]]):
                        ret_dict[a[0]] = [0] * self.array_size
                    ret_dict[a[0]][index:index + padding] = b
                else:
                    entity = a[0]
                    _type = a[1]
                    # Array is empty
                    if not len(ret_dict[entity]):
                        for i in range(self.array_size):
                            ret_dict[entity].append({_type: 0})
                    # _type is not in the array
                    if _type not in ret_dict[entity][0]:
                        for i in range(self.array_size):
                            ret_dict[entity][i][_type] = 0
                    # Set values from transformation to array
                    for i in range(len(b)):
                        ret_dict[entity][index + i][_type] = b[i]
        # We apply a transformation to str on all the values
        return {k: str(ret_dict[k]).replace("'", "\"") for k in ret_dict}



from Supplychain.Generic.folder_io import FolderReader, FolderWriter
from Supplychain.Generic.timer import Timer
from Supplychain.Schema.simulator_files_description import simulator_format
from Supplychain.Schema.default_values import variables_default_values
from Supplychain.Generic.duration import seconds_to_iso_8061_duration
from datetime import datetime, tzinfo, timedelta
from Supplychain.Wrappers.environment_variables import SECONDS_IN_MINUTE

from collections import defaultdict, deque


class simple_utc(tzinfo):
    def tzname(self, **kwargs):
        return "UTC"

    def utcoffset(self, dt):
        return timedelta(0)


def convert_dict(key_type, value_type, dict_item) -> dict:
    if dict_item is None:
        return dict()
    return {key_type(k): value_type(v)
            for k, v
            in sorted(dict_item.items(), key=lambda i: key_type(i[0]))}


def find_effective_value_for_time_step(time_step: int,
                                       schedule: dict) -> float:
    if time_step in schedule:
        return schedule[time_step]

    effective_time_step = max(int(k) for k in schedule.keys() if int(k) <= time_step)
    return schedule.get(str(effective_time_step), 0.0)


class FromDictToSimulatorConverter(Timer):

    def __init__(self,
                 reader: FolderReader,
                 writer: FolderWriter):
        Timer.__init__(self, prefix="[Convert]")

        self.reader = reader
        self.writer = writer

        self.configuration = self.reader.files.get('Configuration', [])[0]

        self.total_time_steps = self.configuration['StepsPerCycle'] * self.configuration['SimulatedCycles']

        self.files_content = dict()

        self.current_group = 0

        for key in simulator_format.keys():
            self.files_content[key] = list()

        self.entity_order = dict()

        self.lead_times = {}

    def __scheduler(self):
        inputs = self.reader.files.get('input', [])
        outputs = self.reader.files.get('output', [])
        contains = self.reader.files.get('contains', [])
        transports = self.reader.files.get('Transport', [])
        stocks = self.reader.files.get('Stock', [])

        operations_names = set(i['target'] for i in inputs)
        transports_names = set(t['Label'] for t in
                               transports)  # TODO to be replaced by RelationshipId when available from azure-digital-twins-simulator-connector

        stocks_names = set(stock['id'] for stock in stocks)
        _input_stocks = set(i['source'] for i in inputs) | set(t['source'] for t in transports)
        _output_stocks = set(o['target'] for o in outputs) | set(t['target'] for t in transports)

        start_stocks = _input_stocks - _output_stocks
        final_stocks = _output_stocks - _input_stocks

        scheduled_elements = []
        for stock in sorted(start_stocks):
            scheduled_element = dict(Type='Rule',
                                     Name=stock,
                                     TaskName='Dispatch')
            scheduled_elements.append(scheduled_element)
        self.files_content['IndustrialNetworkScheduler'].extend(scheduled_elements)

        stocks_queue = deque()

        contractors = set(op['id']
                          for op in self.reader.files.get('ProductionOperation', [])
                          if op['IsContractor'])

        for st in final_stocks:
            stocks_queue.append(st)
            self.entity_order[st] = 0

        parents = defaultdict(set)

        # While we have stocks to treat we continue
        for links in [inputs, outputs]:
            for link in links:
                parents[link['source']].add(link['target'])
        for tr in transports:
            parents[tr['source']].add(tr['Label'])
            parents[tr['Label']].add(tr['target'])

        for production_resource in [c['source'] for c in contains]:
            brotherhood = [c['target'] for c in contains if c['source'] == production_resource]
            first_child = brotherhood.pop()
            for child in brotherhood:
                parents[first_child] |= parents[child]
            for child in brotherhood:
                parents[child] = parents[first_child]

        max_depth = 0

        def get_entity_order(_entity):
            if _entity not in self.entity_order:
                self.entity_order[_entity] = 1 + max(map(get_entity_order, parents[_entity]))
            return self.entity_order[_entity]

        for entity in parents.keys():
            o = get_entity_order(entity)
            max_depth = max(max_depth, o)

        operation_order = [[] for _ in range(max_depth)]
        for k, v in self.entity_order.items():
            if k in transports_names or k in operations_names:
                operation_order[v].append(k)

        operation_order.reverse()
        operation_order = [i for i in operation_order if len(i) > 0]

        scheduled_ComputeStockToTransfer_stocks = set()

        for list_operations in operation_order:
            # For each rank of operation, every operation can be done in parallel, so order is not important

            # To schedule we need some entities and information about them:

            fixed_ee = set()
            # StockToOperation edge entities linked to the contractor Operation
            # entites considered at this rank.

            prop_ee = set()
            # StockToOperation edge entities linked to the regular Operation
            # entites considered at this rank.

            all_fixed_ee = set()
            # StockToOperation edge entities linked to the Stock entities which
            # are taken as input by the contractor Operation entities considered
            #  at this rank and for which ComputeStockToTransfer processes were
            # not scheduled yet.

            all_prop_ee = set()
            # StockToOperation edge entities linked to the Stock entities which
            #  are taken as input by the regular Operation entities considered at
            # this rank and for which ComputeStockToTransfer processes were not
            # scheduled yet.

            input_stocks = set()
            # Stock entities taken as input by the Operation entities considered
            # at this rank and for which ComputeStockToTransfer processes were
            # not scheduled yet.

            machines = set()
            # Machine entities containing the non transport Operation entities
            #  considered at this rank.

            operation_ee = set()
            # StockToOperation edge entities for the non transport Operation
            # entities considered at this rank.

            transport_ops = set()
            # Operation entities considered at this rank which are transports.

            end_stocks = set()
            # Stock entities outputed by the Operation entities considered at
            # this rank.

            for operation in list_operations:
                if operation in operations_names:
                    edge_entities = set(i['source'] + '_to_' + i['target']
                                        for i in inputs
                                        if i['target'] == operation)
                    machines.add(next(c['source']
                                      for c in contains
                                      if c['target'] == operation))
                    stock = next(o['target']
                                 for o in outputs
                                 if o['source'] == operation)
                    operation_ee.update(edge_entities)
                    input_stocks |= set((i['source']
                                         for i in inputs
                                         if i['target'] == operation))
                else:
                    edge_entities = set(t['source'] + '_to_' + t['Label']
                                        for t in transports
                                        if t['Label'] == operation)
                    transport_ops.add(operation)
                    stock = next(t['target']
                                 for t in transports
                                 if t['Label'] == operation)
                    input_stocks |= set((t['source']
                                         for t in transports
                                         if t['Label'] == operation))

                if operation in contractors:
                    fixed_ee.update(edge_entities)
                else:
                    prop_ee.update(edge_entities)

                end_stocks.add(stock)

            input_stocks -= scheduled_ComputeStockToTransfer_stocks
            if input_stocks:
                scheduled_ComputeStockToTransfer_stocks |= input_stocks
                for i in inputs:
                    if i['source'] in input_stocks:
                        operation = i['target']
                        (all_fixed_ee if operation in contractors else all_prop_ee).add(
                            i['source'] + '_to_' + operation)
                for t in transports:
                    if t['source'] in input_stocks:
                        operation = t['Label']
                        (all_fixed_ee if operation in contractors else all_prop_ee).add(
                            t['source'] + '_to_' + operation)

            # Names must be sorted to stabilize the output schedule.
            all_fixed_ee = sorted(all_fixed_ee)
            fixed_ee = sorted(fixed_ee)
            all_prop_ee = sorted(all_prop_ee)
            prop_ee = sorted(prop_ee)
            machines = sorted(machines)
            operation_ee = sorted(operation_ee)
            transport_ops = sorted(transport_ops)
            end_stocks = sorted(end_stocks)

            # Now we can create our scheduled Processes / Rules
            scheduled_elements = []

            # As soon as a ComputeStockToTransfer process is scheduled for a
            # particular stock, all ComputeStockToTransfer processes for the
            # said stock should be scheduled at the same time. Moreover, they
            # should be scheduled only once.
            for EE in all_fixed_ee:
                scheduled_element = dict(Type='Process',
                                         Name=EE,
                                         TaskName='ComputeStockToTransfer')
                scheduled_elements.append(scheduled_element)

            for EE in fixed_ee:
                scheduled_element = dict(Type='Process',
                                         Name=EE,
                                         TaskName='TransferStock')
                scheduled_elements.append(scheduled_element)

            for EE in all_prop_ee:
                scheduled_element = dict(Type='Process',
                                         Name=EE,
                                         TaskName='ComputeStockToTransfer')
                scheduled_elements.append(scheduled_element)

            for EE in prop_ee:
                scheduled_element = dict(Type='Process',
                                         Name=EE,
                                         TaskName='TransferStock')
                scheduled_elements.append(scheduled_element)

            for mac in machines:
                scheduled_element = dict(Type='Rule',
                                         Name=mac,
                                         TaskName='SelectStock')
                scheduled_elements.append(scheduled_element)

            for EE in operation_ee:
                scheduled_element = dict(Type='Process',
                                         Name=EE,
                                         TaskName='SendBackStockIfNecessary')
                scheduled_elements.append(scheduled_element)

            for mac in machines:
                scheduled_element = dict(Type='Process',
                                         Name=mac,
                                         TaskName='LaunchProcessing')
                scheduled_elements.append(scheduled_element)

            for transport in transport_ops:
                scheduled_element = dict(Type='Rule',
                                         Name=transport,
                                         TaskName='SelectStock')
                scheduled_elements.append(scheduled_element)

            for transport in transport_ops:
                scheduled_element = dict(Type='Process',
                                         Name=transport,
                                         TaskName='Process')
                scheduled_elements.append(scheduled_element)

            for stock in end_stocks:
                scheduled_element = dict(Type='Rule',
                                         Name=stock,
                                         TaskName='ReceiveStock')
                scheduled_elements.append(scheduled_element)

            for stock in end_stocks:
                scheduled_element = dict(Type='Rule',
                                         Name=stock,
                                         TaskName='Dispatch')
                scheduled_elements.append(scheduled_element)

            self.files_content['IndustrialNetworkScheduler'].extend(scheduled_elements)
        self.files_content['IndustrialNetworkScheduler'].append(dict(
            Type='Process',
            Name='IndustrialNetwork',
            TaskName='MesoResolveDemands'
        ))
        self.files_content['IndustrialNetworkScheduler'].extend((
            dict(
                Type='Rule',
                Name=stock,
                TaskName='CheckLevelAndOrderIfNecessary'
            )
            for stock in sorted(sorted(stocks_names), key=lambda e: self.entity_order.get(e, -1))
        ))
        self.writer.write_from_list(self.files_content['IndustrialNetworkScheduler'], 'IndustrialNetworkScheduler',
                                    ordering_key=None)

    def __lead_times(self):
        inputs = self.reader.files.get('input', [])
        operations = self.reader.files.get('ProductionOperation', [])
        outputs = self.reader.files.get('output', [])
        stocks = self.reader.files.get('Stock', [])
        transports = self.reader.files.get('Transport', [])

        stock_names = set(stock['id'] for stock in stocks)

        non_input_stock_names = (
            stock_names
            - set(i['source'] for i in inputs)
            - set(transport['source'] for transport in transports)
        )
        non_output_stock_names = (
            stock_names
            - set(output['target'] for output in outputs)
            - set(transport['target'] for transport in transports)
        )

        parents = defaultdict(set)
        for links in [inputs, outputs]:
            for link in links:
                parents[link['target']].add(link['source'])
        for transport in transports:
            parents[transport['Label']].add(transport['source'])
            parents[transport['target']].add(transport['Label'])

        durations = {
            stock['id']: 0
            for stock in stocks
        } | {
            operation['id']: operation['Duration']
            for operation in operations
        } | {
            transport['Label']: transport['Duration']
            for transport in transports
        }

        for stock_name in non_output_stock_names:
            self.lead_times[stock_name] = 0

        def get_lead_time(entity):
            if entity not in self.lead_times:
                self.lead_times[entity] = max(get_lead_time(parent) for parent in parents[entity]) + durations[entity]
            return self.lead_times[entity]

        for stock_name in non_input_stock_names:
            get_lead_time(stock_name)

    def __operation(self):

        operations_csv = self.reader.files.get('ProductionOperation', [])
        contains_csv = self.reader.files.get('contains', [])
        outputs = self.reader.files.get('output', [])
        stock_csv = self.reader.files.get('Stock', [])

        for operation in operations_csv:
            parent_resource = next(c['source']
                                   for c in contains_csv
                                   if c['target'] == operation['id'])
            output_stock = next(
                link['target']
                for link in outputs
                if link['source'] == operation['id']
            )
            follows_stock_policy = next(
                stock['StockPolicy'] != 'None'
                for stock in stock_csv
                if stock['id'] == output_stock
            )
            operation_item = {
                "Name": operation['id'],
                "ParentName": parent_resource,
                "OperatingPerformanceSchedule": convert_dict(int,
                                                             float,
                                                             operation['OperatingPerformances']),
                "CycleTimeSchedule": convert_dict(int,
                                                  float,
                                                  operation['CycleTimes']),
                "RejectRateSchedule": convert_dict(int,
                                                   float,
                                                   operation['RejectRates']),
                "OperatingPerformanceUncertaintySchedule": convert_dict(int,
                                                                        float,
                                                                        operation['OperatingPerformanceUncertainties']),
                "ProductionUnitCostSchedule": convert_dict(int,
                                                           float,
                                                           operation['ProductionUnitCosts']),
                "CO2UnitEmissionsSchedule": convert_dict(int,
                                                         float,
                                                         operation['CO2UnitEmissions']),
                "ProductionPlanSchedule": convert_dict(int,
                                                       float,
                                                       operation['QuantitiesToProduce']),
                "InvestmentCost": operation['InvestmentCost'],
                "IsContractor": operation['IsContractor'],
                "Priority": operation['Priority'],
                "SchedulingStep": self.entity_order.get(operation['id'], -1),
                "FollowsStockPolicy": follows_stock_policy,
                "Duration": operation['Duration'],
                "LeadTime": self.lead_times[operation['id']],
            }
            self.files_content['BE_Operation'].append(operation_item)
        self.writer.write_from_list(self.files_content['BE_Operation'], 'BE_Operation', ordering_key='Name')

    def __model_parameters(self):

        production_plan_is_available = all(bool(po['QuantitiesToProduce'])
                                           for po in self.reader.files.get('ProductionOperation', []))
        model_parameters = {
            "TimeStepDuration":
                seconds_to_iso_8061_duration(self.configuration['TimeStepDuration'] * SECONDS_IN_MINUTE),
            "TimeStepPerCycle":
                self.configuration['StepsPerCycle'],
            "NumberOfCycle":
                self.configuration['SimulatedCycles'],
            "StartingDate":
                datetime.fromisoformat(self.configuration['StartingDate']).replace(tzinfo=simple_utc()).isoformat(),
            "ManageBacklogQuantities":
                self.configuration['ManageBacklogQuantities'],
            "OptimizationObjective":
                self.configuration['OptimizationObjective'],
            "ActivateUncertainties":
                self.configuration['ActivateUncertainties'],
            "EmptyObsoleteStocks":
                self.configuration['EmptyObsoleteStocks'],
            "UncertaintiesProbabilityDistribution":
                self.configuration['UncertaintiesProbabilityDistribution'],
            "ActivateVariableMachineOpeningRate":
                self.configuration['ActivateVariableMachineOpeningRate'],
            "ImmobilizedCashRelativeCost":
                self.configuration['FinancialCostOfStock'],
            "BatchSize":
                self.configuration['BatchSize'],
            "ExtraProductionPlanIsAvailable":
                production_plan_is_available,
            "EnforceProductionPlan":
                self.configuration['EnforceProductionPlan'],
            "AllowPartRetention":
                self.configuration['IntermediaryStockDispatchPolicy'] == 'AllowRetention',
        }

        self.files_content['ModelParameters'].append(model_parameters)
        self.writer.write_from_list(self.files_content['ModelParameters'], 'ModelParameters', ordering_key=None)

    def __stock(self):
        stock_csv = self.reader.files.get('Stock', [])
        for stock in stock_csv:
            maximal_stock = stock['MaximalStock']
            stock_item = {"Name": stock['id'],
                          "CurrentStock": stock['InitialStock'],
                          "MinimalStock": stock['MinimalStock'],
                          "MaximalStock": maximal_stock if maximal_stock is not None else -1,
                          "Value": stock['InitialValue'],
                          "PurchasingUnitCost": stock['PurchasingUnitCost'],
                          "CO2UnitEmissions": stock['CO2UnitEmissions'],
                          "UnitIncome": stock['UnitIncome'],
                          "Demand": dict(),
                          "StorageUnitCostSchedule": convert_dict(int, float, stock['StorageUnitCosts']),
                          "IsInfinite": stock['IsInfinite'],
                          "StockPolicy": stock['StockPolicy'],
                          "SourcingPolicy": stock['SourcingPolicy'],
                          "DispatchPolicy": stock['DispatchPolicy'],
                          "ReviewPeriod": stock['ReviewPeriod'],
                          "NextReview": stock['NextReview'],
                          "Advance": stock['Advance'],
                          "LeadTime": self.lead_times[stock['id']],
                          "SalesForecasts": convert_dict(int, float, stock['SalesForecasts']),
                          "OrderPointSchedule": convert_dict(int, float, stock['OrderPoints']),
                          "OrderQuantitySchedule": dict(),
                          "SafetyQuantitySchedule": convert_dict(int, float, stock['SafetyQuantities']),
                          "SchedulingStep": self.entity_order.get(stock['id'], -1)}

            if stock_item['StockPolicy'] == 'OrderPointFixedQuantity':
                stock_item['OrderQuantitySchedule'] = convert_dict(int, float, stock['OrderQuantities'])
            elif stock_item['StockPolicy'] == 'OrderPointVariableQuantity':
                stock_item['OrderQuantitySchedule'] = convert_dict(int, float, stock['OrderUpToLevels'])

            backlog_weight = stock['BacklogWeight']
            last_uncertainty = 0

            if stock['Demands']:
                demands = convert_dict(int, float, stock['Demands'])
                demands_weight = convert_dict(int, float, stock['DemandWeights'])
                demands_uncertainties = convert_dict(int, float, stock['DemandUncertainties'])

                last_weight = 0

                for time_step in range(self.total_time_steps):
                    if time_step in demands_weight:
                        last_weight = demands_weight[time_step]
                    if time_step in demands_uncertainties:
                        last_uncertainty = demands_uncertainties[time_step]
                    if time_step in demands and demands[time_step]:
                        current_demand = dict(ExternalDemand=demands[time_step],
                                              InternalDemand=0.0,
                                              BacklogWeight=backlog_weight,
                                              ExternalWeight=last_weight,
                                              InternalWeight=0.0,
                                              WeightMax=0.0,
                                              MaxVal=0.0,
                                              DemandRelativeUncertainty=last_uncertainty)
                        stock_item['Demand'][time_step] = current_demand

            maximization_weight = stock['MaximizationWeight']
            if maximization_weight:
                total_max_production = 0
                for time_step in range(self.total_time_steps):
                    total_max_production += self.compute_theoretical_maximum_production(stock['id'],
                                                                                      time_step)
                if self.total_time_steps - 1 not in stock_item['Demand']:
                    current_demand = dict(ExternalDemand=0.0,
                                          InternalDemand=0.0,
                                          BacklogWeight=backlog_weight,
                                          ExternalWeight=0.0,
                                          InternalWeight=0.0,
                                          WeightMax=0.0,
                                          MaxVal=0.0,
                                          DemandRelativeUncertainty=last_uncertainty)
                    stock_item['Demand'][self.total_time_steps - 1] = current_demand
                stock_item['Demand'][self.total_time_steps - 1]['WeightMax'] = maximization_weight
                stock_item['Demand'][self.total_time_steps - 1]['MaxVal'] = total_max_production

            self.files_content['BE_Stock'].append(stock_item)
        self.writer.write_from_list(self.files_content['BE_Stock'], 'BE_Stock', ordering_key='Name')

    def compute_theoretical_maximum_production(self, stock_id: str, time_step: int) -> float:
        stock = next(
            s
            for s in self.reader.files.get('Stock', [])
            if s['id'] == stock_id
        )
        theoretical_maximum = stock['InitialStock'] if time_step == 0 else 0.0
        transports = [
            transport
            for transport in self.reader.files.get('Transport', [])
            if transport['target'] == stock_id
        ]
        for transport in transports:
            origin = transport['source']
            duration = transport['Duration']
            if time_step >= duration:
                theoretical_maximum += self.compute_theoretical_maximum_production(origin, time_step - duration)
            else:
                theoretical_maximum += transport['InitialTransportedQuantities'].get(str(duration - time_step - 1), 0.0)
        operation_names = set(
            output['source']
            for output in self.reader.files.get('output', [])
            if output['target'] == stock_id
        )
        operations = [
            operation
            for operation in self.reader.files.get('ProductionOperation', [])
            if operation['id'] in operation_names
        ]
        for operation in operations:
            if operation['IsContractor']:
                theoretical_maximum += find_effective_value_for_time_step(time_step,
                                                                        operation['QuantitiesToProduce'])
            else:
                pr_name = next(
                    contains['source']
                    for contains in self.reader.files.get('contains', [])
                    if contains['target'] == operation['id']
                )
                pr_data = next(
                    pr
                    for pr in self.reader.files.get('ProductionResource', [])
                    if pr['id'] == pr_name
                )
                max_opening_time = find_effective_value_for_time_step(time_step, pr_data['OpeningTimes'])
                ct = (
                    find_effective_value_for_time_step(time_step, operation['CycleTimes'])
                    if operation['CycleTimes']
                    else variables_default_values['ProductionOperation']['CycleTimes']
                )
                rr = (
                    find_effective_value_for_time_step(time_step, operation['RejectRates'])
                    if operation['RejectRates']
                    else variables_default_values['ProductionOperation']['RejectRates']
                )
                op = (
                    find_effective_value_for_time_step(time_step, operation['OperatingPerformances'])
                    if operation['OperatingPerformances']
                    else variables_default_values['ProductionOperation']['OperatingPerformances']
                )
                theoretical_maximum += (max_opening_time / ct) * op * (1.0 - rr)
        return theoretical_maximum

    def __machine(self):
        pr_csv = self.reader.files.get('ProductionResource', [])
        contains_csv = self.reader.files.get('contains', [])

        for pr in pr_csv:
            target_operations = [c['target']
                                 for c in contains_csv
                                 if c['source'] == pr['id']]
            operations_count = len(target_operations)
            target_propotions = {op: 1.0 / operations_count
                                 for op in sorted(target_operations)}

            group = {}
            if operations_count > 1:
                group = {self.current_group: {"StartCycle": 0,
                                              "EndCycle": self.configuration['SimulatedCycles'] - 1}}
                self.current_group += 1

            pr_item = {"Name": pr['id'],
                       "StockTypeProportion": [target_propotions] * self.configuration['SimulatedCycles'],
                       "Groups": group,
                       "OpeningRates": [1] * self.configuration['SimulatedCycles'],
                       "OpeningTimeSchedule": convert_dict(int,
                                                           float,
                                                           pr['OpeningTimes']),
                       "ProductionCostSchedule": convert_dict(int,
                                                              float,
                                                              pr['FixedProductionCosts']),
                       "ProductionPolicy": pr['ProductionPolicy']}
            self.files_content['CE_Machine'].append(pr_item)
        self.writer.write_from_list(self.files_content['CE_Machine'], 'CE_Machine', ordering_key='Name')

    def __op_to_stock(self):
        output_csv = self.reader.files.get('output', [])
        for o in output_csv:
            op_to_stock_item = {"StockName": o['target'],
                                "OperationName": o['source']}
            self.files_content['Arcs_OperationToStock'].append(op_to_stock_item)
        output_csv = self.reader.files.get('Transport', [])
        for o in output_csv:
            op_to_stock_item = {"StockName": o['target'],
                                "OperationName": o['Label']}
            self.files_content['Arcs_OperationToStock'].append(op_to_stock_item)
        self.writer.write_from_list(self.files_content['Arcs_OperationToStock'], 'Arcs_OperationToStock',
                                    ordering_key='StockName')

    def __transport(self):
        transport_csv = self.reader.files.get('Transport', [])
        stock_csv = self.reader.files.get('Stock', [])

        def dict_to_list(duration: int, data: dict):
            if data:
                return [data.get(str(t), 0.0) for t in range(duration)]
            return [0.0 for _ in range(duration)]

        for transport in transport_csv:
            follows_stock_policy = next(
                stock['StockPolicy'] != 'None'
                for stock in stock_csv
                if stock['id'] == transport['target']
            )
            transport_item = {
                "Name": transport['Label'],
                # TODO to be replaced by RelationshipId when available from azure-digital-twins-simulator-connector
                "Duration": transport['Duration'],
                "InitialTransportQuantity": dict_to_list(transport['Duration'],
                                                         transport['InitialTransportedQuantities']),
                "InitialTransportValue": dict_to_list(transport['Duration'],
                                                      transport['InitialTransportedValues']),
                "TransportUnitCostSchedule": convert_dict(int,
                                                          float,
                                                          transport['TransportUnitCosts']),
                "CustomsUnitCostSchedule": convert_dict(int,
                                                        float,
                                                        transport['CustomFees']),
                "CO2UnitEmissionsSchedule": convert_dict(int,
                                                         float,
                                                         transport['CO2UnitEmissions']),
                "Priority": transport['Priority'],
                "SchedulingStep": self.entity_order.get(transport['Label'], -1),
                "FollowsStockPolicy": follows_stock_policy,
                "LeadTime": self.lead_times[transport['Label']],
            }
            self.files_content['BE_TransportOperation'].append(transport_item)

        self.writer.write_from_list(self.files_content['BE_TransportOperation'], 'BE_TransportOperation',
                                    ordering_key='Name')

    def __stock_to_operation(self):
        input_csv = self.reader.files.get('input', [])
        transport_csv = self.reader.files.get('Transport', [])

        number_of_edges_by_stock = defaultdict(int)

        for i in input_csv:
            number_of_edges_by_stock[i['source']] += 1

        for t in transport_csv:
            number_of_edges_by_stock[t['source']] += 1

        groups = dict()

        stock_has_demand = {s['id']: bool(s['Demands'])  # DS
                            for s in self.reader.files.get('Stock', [])}

        for stock in sorted(number_of_edges_by_stock.keys()):
            groups[stock] = {}
            if number_of_edges_by_stock[stock] > 1 or stock_has_demand[stock]:
                groups[stock] = {self.current_group: {"StartCycle": 0,
                                                      "EndCycle": self.configuration['SimulatedCycles'] - 1}}
                self.current_group += 1

        contractors = set(op['id']
                          for op in self.reader.files.get('ProductionOperation', [])
                          if op['IsContractor'])

        for i in input_csv:
            prop_sto = [1 / number_of_edges_by_stock[i['source']]] * self.configuration['SimulatedCycles']
            sto_item = {"Name": i['source'] + "_to_" + i['target'],
                        "OperationName": i['target'],
                        "StockName": i['source'],
                        "PropStockForOperation": prop_sto,
                        "IsProportion": i['target'] not in contractors,
                        "Groups": groups[i['source']],
                        "Cardinality": i['InputQuantity']}
            self.files_content['EE_StockToOperation'].append(sto_item)

        for i in transport_csv:
            prop_sto = [1 / number_of_edges_by_stock[i['source']]] * self.configuration['SimulatedCycles']
            sto_item = {"Name": i['source'] + "_to_" + i['Label'],
                        "OperationName": i['Label'],
                        # TODO to be replaced by RelationshipId when available from azure-digital-twins-simulator-connector
                        "StockName": i['source'],
                        "PropStockForOperation": prop_sto,
                        "IsProportion": True,
                        "Groups": groups[i['source']],
                        "Cardinality": 1}
            self.files_content['EE_StockToOperation'].append(sto_item)

        self.writer.write_from_list(self.files_content['EE_StockToOperation'], 'EE_StockToOperation',
                                    ordering_key='Name')

    def transform(self):
        self.__scheduler()
        self.split("{time_since_last_split:6.4f} ({time_since_start:6.4f}) - IndustrialNetworkScheduler")
        self.__lead_times()
        self.split("{time_since_last_split:6.4f} ({time_since_start:6.4f}) - Lead times")
        self.__stock()
        self.split("{time_since_last_split:6.4f} ({time_since_start:6.4f}) - BE_Stock")
        self.__operation()
        self.split("{time_since_last_split:6.4f} ({time_since_start:6.4f}) - BE_Operation")
        self.__transport()
        self.split("{time_since_last_split:6.4f} ({time_since_start:6.4f}) - BE_TransportOperation")
        self.__machine()
        self.split("{time_since_last_split:6.4f} ({time_since_start:6.4f}) - CE_Machine")
        self.__op_to_stock()
        self.split("{time_since_last_split:6.4f} ({time_since_start:6.4f}) - Arcs_OperationToStock")
        self.__stock_to_operation()
        self.split("{time_since_last_split:6.4f} ({time_since_start:6.4f}) - EE_StockToOperation")
        self.__model_parameters()
        self.split("{time_since_last_split:6.4f} ({time_since_start:6.4f}) - ModelParameters")



import os
import uuid

if "CSM_SIMULATION_ID" not in os.environ:
    os.environ["CSM_SIMULATION_ID"] = str(uuid.uuid1())

SECONDS_IN_MINUTE = 60


def get_alternative_environ_var(first_key: str, second_key: str, default_value: str):
    if first_key in os.environ:
        return os.environ[first_key]
    return os.environ.get(second_key, default_value)


class EnvironmentVariables:
    from_adt_folder = get_alternative_environ_var('SUPPLY_DRIVER_FROM_ADT_FOLDER',
                                                  'CSM_DATASET_ABSOLUTE_PATH',
                                                  '/mnt/scenariorun-data/')
    parameters_folder = get_alternative_environ_var('SUPPLY_DRIVER_PARAMETER_FOLDER',
                                                    'CSM_PARAMETERS_ABSOLUTE_PATH',
                                                    '/mnt/scenariorun-parameters/')
    temp_folder = os.environ.get('SUPPLY_DRIVER_TEMP_FOLDER',
                                 '/tmp/supply_tmp/')
    simulation_import_folder = os.environ.get('SUPPLY_DRIVER_SIMU_IMPORT_FOLDER',
                                              '/pkg/share/Simulation/Resource/Import')
    simulation_id = os.environ.get('CSM_SIMULATION_ID', str(uuid.uuid1()))
    simulation_name = os.environ.get('CSM_SIMULATION_VAR', "Simulation")
    amqp_consumer = os.environ.get('CSM_PROBES_MEASURES_TOPIC', None)
    adx_parameters = {
        "uri": os.environ.get('AZURE_DATA_EXPLORER_RESOURCE_URI', None),
        "ingest-uri": os.environ.get('AZURE_DATA_EXPLORER_RESOURCE_INGEST_URI', None),
        "database": os.environ.get('AZURE_DATA_EXPLORER_DATABASE_NAME', None)
    }



class ADTColumnDescription:
    """
    This class contains a descriptor of the format of a dataset in ADT
    It allows the description of each column and how it should be considered during a transformation
    - fixed : means that the data is a non changing value that stay constant through time,
      or can be used as an initialization
    - change : represent a change over time which means it needs a timestep for when it takes effect,
      and the new value for the subsequent time steps
    - event : represent a ponctual event over time, the change is not permanent and on subsequent timestep
      without other values nothing happens
    """
    format = {
        "Stock": {
            "fixed": [
                "id",
                "Label",
                "PlantName",
                "Step",
                "PartId",
                "MinimalStock",
                "MaximalStock",
                "InitialStock",
                "InitialValue",
                "IsInfinite",
                "PurchasingUnitCost",
                "CO2UnitEmissions",
                "UnitIncome",
                "BacklogWeight",
                "MaximizationWeight",
                "StockPolicy",
                "SourcingPolicy",
                "DispatchPolicy",
                "ReviewPeriod",
                "NextReview",
                "Advance",
            ],
            "change": [
                "StorageUnitCosts",
                "OrderPoints",
                "OrderQuantities",
                "OrderUpToLevels",
                "SafetyQuantities",
            ],
            "event": {
                "Demands": [
                    "Demands",
                    "DemandUncertainties",
                    "DemandWeights",
                ],
                "SalesForecasts": [
                    "SalesForecasts",
                ],
            }
        },
        "Transport": {
            "fixed": [
                "Label",  # TODO to be replaced by RelationshipId when available from azure-digital-twins-simulator-connector
                "source",
                "target",
                "Duration",
                "Priority",
            ],
            "change": [
                "CustomFees",
                "TransportUnitCosts",
                "CO2UnitEmissions",
            ],
            "event": {
                "InitialTransports": [
                    "InitialTransportedQuantities",
                    "InitialTransportedValues",
                ]
            },
        },
        "Configuration": {
            "fixed": [
                "ActivateUncertainties",
                "ActivateVariableMachineOpeningRate",
                "BatchSize",
                "EmptyObsoleteStocks",
                "EnforceProductionPlan",
                "FinancialCostOfStock",
                "ManageBacklogQuantities",
                "OptimizationObjective",
                "SimulatedCycles",
                "StartingDate",
                "StepsPerCycle",
                "TimeStepDuration",
                "UncertaintiesProbabilityDistribution",
                "IntermediaryStockDispatchPolicy",
            ],
            "change": [],
            "event": {}
        },
        "contains": {
            "fixed": [
                "source",
                "target",
            ],
            "change": [],
            "event": {}
        },
        "output": {
            "fixed": [
                "source",
                "target",
            ],
            "change": [],
            "event": {}
        },
        "input": {
            "fixed": [
                "source",
                "target",
                "InputQuantity",
            ],
            "change": [],
            "event": {}
        },
        "ProductionResource": {
            "fixed": [
                "id",
                "Label",
                "PlantName",
                "ProductionStep",
                "ProductionPolicy",
            ],
            "change": [
                "FixedProductionCosts",
                "OpeningTimes",
            ],
            "event": {}
        },
        "ProductionOperation": {
            "fixed": [
                "id",
                "Label",
                "PlantName",
                "IsContractor",
                "InvestmentCost",
                "Priority",
                "Duration",
            ],
            "change": [
                "QuantitiesToProduce",
                "OperatingPerformances",
                "CycleTimes",
                "RejectRates",
                "OperatingPerformanceUncertainties",
                "ProductionUnitCosts",
                "CO2UnitEmissions",
            ],
            "event": {}
        }
    }



import pandas as pd
import numpy as np
from Supplychain.Generic.folder_io import FolderReader, FolderWriter
from Supplychain.Schema.adt_column_description import ADTColumnDescription
from Supplychain.Generic.timer import Timer


def write_transformed_data(reader: FolderReader, writer: FolderWriter):
    with FromTableToDictConverter(reader, writer) as converter:
        return converter.convert_all()


dataset_description = ADTColumnDescription.format


class FromTableToDictConverter(Timer):

    def __init__(self, reader: FolderReader, writer: FolderWriter, descriptor=dataset_description):
        Timer.__init__(self, "[Table -> Dict]")
        self.descriptor = descriptor
        self.reader = reader
        self.writer = writer

    def check_indices(self, name: str, indices: list):
        errors = 0
        unique_indices = set(indices)
        if len(unique_indices) < len(indices):
            duplicates = [
                (index, count)
                for index, count in zip(
                    unique_indices,
                    (indices.count(index) for index in unique_indices),
                )
                if count > 1
            ]
            n = len(duplicates)
            errors += n
            self.display_message(f"ERROR: {name} value{'s' if n > 1 else ''} defined multiple times:")
            for index, count in duplicates:
                self.display_message(f"  {', '.join(str(i) for i in index) if isinstance(index, tuple) else index}: {count}")
        return errors

    def convert_file(self, file_name: str):
        errors = 0

        file_description = dataset_description[file_name]

        entities = []
        if file_name not in self.reader.files:
            return errors

        fixed_file_to_convert = self.reader.files[file_name]

        master_key = None
        if 'id' in file_description['fixed']:
            master_key = 'id'
        elif 'Label' in file_description['fixed']:
            master_key = 'Label'

        for element in fixed_file_to_convert:
            entity = {k: v for k, v in element.items() if k in file_description['fixed']}
            extra = {k: v for k, v in element.items() if k not in file_description['fixed']}
            if extra:
                entity['extra'] = extra
            if master_key and entity.get(master_key) is None:
                continue
            entities.append(entity)

        if master_key:
            errors += self.check_indices(file_name, [entity[master_key] for entity in entities])

        if file_description['event'] or file_description['change']:

            entity_by_id = dict()
            for entity in entities:
                entity_by_id[entity[master_key]] = entity

            time_files = dict()
            time_files[file_name+"Schedules"] = file_description['change']
            for _event_name, _event_descriptor in file_description['event'].items():
                time_files[_event_name] = _event_descriptor

            for _sheet_name, _columns in time_files.items():
                if _sheet_name not in self.reader.files or not _columns:
                    continue
                _sheet_to_convert = self.reader.files[_sheet_name]
                indices = []
                for line in _sheet_to_convert:
                    _id = line.get(master_key)
                    if _id is None:
                        continue
                    timestep = str(line.get('Timestep') or 0)
                    indices.append((_id, timestep))
                    target_entity = entity_by_id[_id]
                    for column in (c for c in _columns if line.get(c) is not None):
                        if column not in target_entity:
                            target_entity[column] = dict()
                        if column in line:
                            target_entity[column][timestep] = line[column]
                errors += self.check_indices(file_name, indices)

        for entity in entities:
            for change in file_description['change']:
                raw_values = entity.get(change)
                if not raw_values:
                    continue
                timesteps = (t for t in sorted(raw_values.keys(), key=int))
                last_timestep = next(timesteps)
                filtered_values = {last_timestep: raw_values[last_timestep]}
                for timestep in timesteps:
                    if raw_values[timestep] != filtered_values[last_timestep]:
                        filtered_values[timestep] = raw_values[timestep]
                        last_timestep = timestep
                entity[change] = filtered_values

        if entities:
            self.writer.write_from_list(entities, file_name, master_key)
        return errors

    def convert_all(self):
        errors = 0
        for _entity_type in self.descriptor.keys():
            self.display_message(f"Converting {_entity_type}")
            errors += self.convert_file(_entity_type)
            self.display_message(f"Converting {_entity_type} : Done")
        self.display_message("File is fully converted")
        return errors



from Supplychain.Protocol.protocol import AbstractOptimization
from Supplychain.Protocol.protocol import AbstractOptimizationAlgorithm
from Supplychain.Protocol.protocol import AbstractObjectiveFunction
from Supplychain.Protocol.protocol import AbstractParameterTransformation
from Supplychain.Generic.timer import Timer
from Supplychain.Wrappers.simulator import CosmoEngine

import multiprocessing
import numpy as np
from typing import Union


class MultiprocessingOptimization(AbstractOptimization, Timer):

    def __init__(self,
                 optimization_algorithm: AbstractOptimizationAlgorithm,
                 objective_function: AbstractObjectiveFunction,
                 parameter_transformation: AbstractParameterTransformation,
                 simulation_name: str,
                 population_size: int,
                 amqp_consumer_adress: Union[str, None] = None):
        Timer.__init__(self, "[MultiprocessingOptimization]")
        AbstractOptimization.__init__(self,
                                      optimization_algorithm,
                                      objective_function,
                                      parameter_transformation,
                                      None)
        self.use_amqp = amqp_consumer_adress is not None
        self.amqp_consumer_adress = amqp_consumer_adress

        self.simulation_name = simulation_name
        self.population_size = population_size

    def run_simulation(self,
                       dv_input: dict):
        # Remove logs other than errors for the simulator
        _logger = CosmoEngine.LoggerManager_GetInstance().GetLogger()
        _logger.SetLogLevel(_logger.eAssertion)
        _simulator = CosmoEngine.LoadSimulator('Simulation')

        # Applies the modification to the simulator
        for datapath, stringvalue in dv_input.items():
            _simulator.FindAttribute(datapath).SetAsString(stringvalue)

        # Add a random seed in case uncertainties are activated
        _simulator.FindAttribute("{Model}Model::{Attribute}Seed").SetAsString(
            str(int(np.random.randint(np.iinfo(np.int32).max)))
        )

        # Destroy all old consumers
        for consumer in _simulator.GetConsumers():
            _simulator.DestroyConsumer(consumer)

        # Define a custom consumer for the optimization
        class PerformanceConsumer(CosmoEngine.Consumer):
            data = list()

            def Consume(self, p_data):
                f = CosmoEngine.PerformanceIndicatorsProbeOutput.Cast(p_data).GetFacts()
                for fact in f:
                    self.data.append({
                        "Profit": fact.GetProfit().GetAsFloat(),
                        "ServiceLevelIndicator": fact.GetServiceLevelIndicator().GetAsFloat()
                    })

        # Instantiate and link the new consumer to the model
        local_consumer = PerformanceConsumer()
        local_consumer.SetName("PerformanceConsumer")
        local_consumer.AddProbe("PerformanceIndicators")
        _simulator.AddConsumer(local_consumer)

        # Run the simulation
        _simulator.Run()

        # return the objective value
        return self.obj_func.applies(local_consumer.data)

    def optimize_parameters(self,
                            is_uncertainty_analysis=False):

        # During initial tests SLO found that forkserver was a viable solution for a multiprocess optim environment
        multiprocessing.set_start_method('forkserver')

        # Define a pool of processes
        self.display_message("Starting Optimization")
        with multiprocessing.Pool(self.population_size) as p:
            # Run the optim algorithm until it has the wanted results
            while self.opt_algo.is_not_finished():
                # Define a list of parameters to be send to the simulation
                parameters = self.opt_algo.generate_decision_variables()
                _params = [(self.par_tran.applies(x.tolist(), 0),)
                           for x in parameters]
                # Run the multiple simulations in the processes pool
                results = p.starmap(self.run_simulation, _params)
                # Update the algorithm
                self.opt_algo.update_algorithm(parameters, results)
        self.display_message("Optimization finished in {time_since_start}")
